# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
E1300 - Module for testing the 11g and 12c verifier types. No special
configuration is needed but a user is created and dropped.
"""

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    requires_connection = False
    user = "ext_test_1300_user"

    @classmethod
    def setUpClass(cls):
        cls.admin_conn = test_env.get_admin_connection()
        cls.password = test_env.get_random_string()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(
                "select count(*) from user$ where name = :1",
                [cls.user.upper()],
            )
            (count,) = cursor.fetchone()
            keyword = "create" if count == 0 else "alter"
            cursor.execute(
                f"{keyword} user {cls.user} identified by {cls.password}"
            )
            cursor.execute(
                "select spare4 from user$ where name = :1", [cls.user.upper()]
            )
            (password_data,) = cursor.fetchone()
            cls.verifier_11g, cls.verifier_12c = password_data.split(";")
            cursor.execute(f"drop user {cls.user}")

    @classmethod
    def tearDownClass(cls):
        user = test_env.get_main_user()
        password = test_env.get_main_password()
        with cls.admin_conn.cursor() as cursor:
            cursor.execute(f"alter user {user} identified by {password}")

    def _verify_connection(self, verifier):
        """
        Verify the ability to connect to the database using the given verifier.
        """
        user = test_env.get_main_user()
        sql = f"alter user {user} identified by values '{verifier}'"
        with self.admin_conn.cursor() as cursor:
            cursor.execute(sql)
        conn = oracledb.connect(
            user=user,
            password=self.password,
            dsn=test_env.get_connect_string(),
        )
        with conn.cursor() as cursor:
            cursor.execute("select user from dual")
            (fetched_user,) = cursor.fetchone()
            self.assertEqual(fetched_user, user.upper())

    def test_ext_1300(self):
        "E1300 - test with an 11g verifier"
        self._verify_connection(self.verifier_11g)

    def test_ext_1301(self):
        "E1301 - test with a 12c verifier"
        self._verify_connection(self.verifier_12c)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
2400 - Module for testing pools
"""

import threading
import unittest

import oracledb
import test_env


class TestCase(test_env.BaseTestCase):
    require_connection = False

    def __connect_and_drop(self):
        with self.pool.acquire() as conn:
            cursor = conn.cursor()
            cursor.execute("select count(*) from TestNumbers")
            (count,) = cursor.fetchone()
            self.assertEqual(count, 10)

    def __connect_and_generate_error(self):
        with self.pool.acquire() as conn:
            cursor = conn.cursor()
            with self.assertRaisesFullCode("ORA-01476"):
                cursor.execute("select 1 / 0 from dual")

    def __callable_session_callback(self, conn, requested_tag):
        self.session_called = True

        supported_formats = {
            "SIMPLE": "'YYYY-MM-DD HH24:MI'",
            "FULL": "'YYYY-MM-DD HH24:MI:SS'",
        }

        supported_time_zones = {"UTC": "'UTC'", "MST": "'-07:00'"}

        supported_keys = {
            "NLS_DATE_FORMAT": supported_formats,
            "TIME_ZONE": supported_time_zones,
        }
        if requested_tag is not None:
            state_parts = []
            for directive in requested_tag.split(";"):
                parts = directive.split("=")
                if len(parts) != 2:
                    raise ValueError("Tag must contain key=value pairs")
                key, value = parts
                value_dict = supported_keys.get(key)
                if value_dict is None:
                    raise ValueError(
                        "Tag only supports keys: %s"
                        % (", ".join(supported_keys))
                    )
                actual_value = value_dict.get(value)
                if actual_value is None:
                    raise ValueError(
                        "Key %s only supports values: %s"
                        % (key, ", ".join(value_dict))
                    )
                state_parts.append(f"{key} = {actual_value}")
            sql = f"alter session set {' '.join(state_parts)}"
            cursor = conn.cursor()
            cursor.execute(sql)
        conn.tag = requested_tag

    def __perform_reconfigure_test(
        self,
        parameter_name,
        parameter_value,
        min=3,
        max=30,
        increment=4,
        timeout=5,
        wait_timeout=5000,
        stmtcachesize=25,
        max_lifetime_session=1000,
        max_sessions_per_shard=3,
        ping_interval=30,
        getmode=oracledb.POOL_GETMODE_WAIT,
        soda_metadata_cache=False,
    ):
        creation_args = dict(
            min=min,
            max=max,
            increment=increment,
            timeout=timeout,
            stmtcachesize=stmtcachesize,
            ping_interval=ping_interval,
            getmode=getmode,
        )
        if test_env.get_client_version() >= (12, 1):
            creation_args["max_lifetime_session"] = max_lifetime_session
        if test_env.get_client_version() >= (12, 2):
            creation_args["wait_timeout"] = wait_timeout
        if test_env.get_client_version() >= (18, 3):
            creation_args["max_sessions_per_shard"] = max_sessions_per_shard
        if test_env.get_client_version() >= (19, 11):
            creation_args["soda_metadata_cache"] = soda_metadata_cache

        pool = test_env.get_pool(**creation_args)
        conn = pool.acquire()

        reconfigure_args = {}
        reconfigure_args[parameter_name] = parameter_value
        pool.reconfigure(**reconfigure_args)
        conn.close()

        actual_args = {}
        for name in creation_args:
            actual_args[name] = getattr(pool, name)
        expected_args = creation_args.copy()
        expected_args.update(reconfigure_args)
        self.assertEqual(actual_args, expected_args)

    def __verify_connection(
        self, connection, expected_user, expected_proxy_user=None
    ):
        cursor = connection.cursor()
        cursor.execute(
            """
            select
                sys_context('userenv', 'session_user'),
                sys_context('userenv', 'proxy_user')
            from dual
            """
        )
        actual_user, actual_proxy_user = cursor.fetchone()
        self.assertEqual(actual_user, expected_user.upper())
        self.assertEqual(
            actual_proxy_user,
            expected_proxy_user and expected_proxy_user.upper(),
        )

    def __verify_create_arg(self, arg_name, arg_value, sql):
        args = {}
        args[arg_name] = arg_value
        pool = test_env.get_pool(**args)
        with pool.acquire() as conn:
            cursor = conn.cursor()
            cursor.execute(sql)
            (fetched_value,) = cursor.fetchone()
            self.assertEqual(fetched_value, arg_value)
        pool.close()

    @unittest.skipIf(
        test_env.get_client_version() < (19, 1), "not supported on this client"
    )
    def test_2400(self):
        "2400 - test getting default pool parameters"
        pool = test_env.get_pool()
        self.assertEqual(pool.busy, 0)
        self.assertEqual(pool.dsn, test_env.get_connect_string())
        self.assertEqual(pool.tnsentry, pool.dsn)
        if test_env.get_client_version() >= (12, 2):
            self.assertEqual(pool.getmode, oracledb.POOL_GETMODE_WAIT)
            self.assertIs(pool.getmode, oracledb.PoolGetMode.WAIT)
        self.assertTrue(pool.homogeneous)
        self.assertEqual(pool.increment, 1)
        self.assertEqual(pool.max, 2)
        if test_env.get_client_version() >= (12, 1):
            self.assertEqual(pool.max_lifetime_session, 0)
        if not test_env.get_is_thin() and test_env.get_client_version() >= (
            18,
            3,
        ):
            self.assertEqual(pool.max_sessions_per_shard, 0)
        self.assertEqual(pool.min, 1)
        if test_env.get_is_thin():
            self.assertIsNone(pool.name)
        else:
            self.assertRegex(pool.name, "^OCI:SP:.+")
        self.assertEqual(pool.ping_interval, 60)
        self.assertEqual(pool.stmtcachesize, oracledb.defaults.stmtcachesize)
        if not test_env.get_is_thin() and test_env.get_client_version() >= (
            19,
            11,
        ):
            self.assertFalse(pool.soda_metadata_cache)
        self.assertEqual(pool.thin, test_env.get_is_thin())
        self.assertEqual(pool.timeout, 0)
        self.assertEqual(pool.username, test_env.get_main_user())

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support proxy users yet"
    )
    def test_2401(self):
        "2401 - test that proxy authentication is possible"
        pool = test_env.get_pool(
            min=2, max=8, increment=3, getmode=oracledb.POOL_GETMODE_WAIT
        )
        self.assertTrue(
            pool.homogeneous, "homogeneous should be True by default"
        )
        with self.assertRaisesFullCode("DPI-1012"):
            pool.acquire(user="missing_proxyuser")
        pool = test_env.get_pool(
            min=2,
            max=8,
            increment=3,
            getmode=oracledb.POOL_GETMODE_WAIT,
            homogeneous=False,
        )
        msg = "homogeneous should be False after setting it in the constructor"
        self.assertFalse(pool.homogeneous, msg)
        conn = pool.acquire(user=test_env.get_proxy_user())
        cursor = conn.cursor()
        cursor.execute("select user from dual")
        (user,) = cursor.fetchone()
        self.assertEqual(user, test_env.get_proxy_user().upper())
        conn.close()

    @unittest.skipIf(
        test_env.get_client_version() < (19, 1), "not supported on this client"
    )
    def test_2402(self):
        "2402 - test setting pool attributes"
        pool = test_env.get_pool()
        test_values = [
            ((11, 2), "ping_interval", 30),
            ((11, 2), "stmtcachesize", 100),
            ((11, 2), "timeout", 10),
            ((12, 2), "getmode", oracledb.POOL_GETMODE_TIMEDWAIT),
            ((12, 1), "max_lifetime_session", 3),
        ]
        for version, attr_name, value in test_values:
            if test_env.get_client_version() >= version:
                setattr(pool, attr_name, value)
                self.assertEqual(getattr(pool, attr_name), value)
                self.assertRaises(
                    TypeError, setattr, pool, attr_name, "invalid value"
                )

        if not test_env.get_is_thin() and test_env.get_client_version() >= (
            18,
            3,
        ):
            self.assertEqual(pool.max_sessions_per_shard, 0)
            self.assertRaises(
                TypeError, setattr, pool, "max_sessions_per_shard", "bad_val"
            )

        if not test_env.get_is_thin() and test_env.get_client_version() >= (
            19,
            11,
        ):
            pool.soda_metadata_cache = True
            self.assertTrue(pool.soda_metadata_cache)
            self.assertRaises(
                TypeError, setattr, pool, "soda_metadata_cache", 22
            )

    def test_2403(self):
        "2403 - connection rolls back before released back to the pool"
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_WAIT)
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("truncate table TestTempTable")
        cursor.execute("insert into TestTempTable (IntCol) values (1)")
        cursor.close()
        pool.release(conn)
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_WAIT)
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("select count(*) from TestTempTable")
        (count,) = cursor.fetchone()
        self.assertEqual(count, 0)
        conn.close()

    def test_2404(self):
        "2404 - test session pool with multiple threads"
        self.pool = test_env.get_pool(
            min=5, max=20, increment=2, getmode=oracledb.POOL_GETMODE_WAIT
        )
        threads = []
        for i in range(20):
            thread = threading.Thread(None, self.__connect_and_drop)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

    def test_2405(self):
        "2405 - test session pool with multiple threads (with errors)"
        self.pool = test_env.get_pool(
            min=5, max=20, increment=2, getmode=oracledb.POOL_GETMODE_WAIT
        )
        threads = []
        for i in range(20):
            thread = threading.Thread(None, self.__connect_and_generate_error)
            threads.append(thread)
            thread.start()
        for thread in threads:
            thread.join()

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_2406(self):
        "2406 - test session pool with various types of purity"
        pool = test_env.get_pool(
            min=1, max=8, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )

        # get connection and set the action
        action = "TEST_ACTION"
        conn = pool.acquire()
        conn.action = action
        cursor = conn.cursor()
        cursor.execute("select 1 from dual")
        cursor.close()
        pool.release(conn)
        self.assertEqual(pool.opened, 1, "opened (1)")

        # verify that the connection still has the action set on it
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("select sys_context('userenv', 'action') from dual")
        (result,) = cursor.fetchone()
        self.assertEqual(result, action)
        cursor.close()
        pool.release(conn)
        self.assertEqual(pool.opened, 1, "opened (2)")

        # get a new connection with new purity (should not have state)
        conn = pool.acquire(purity=oracledb.ATTR_PURITY_NEW)
        cursor = conn.cursor()
        cursor.execute("select sys_context('userenv', 'action') from dual")
        (result,) = cursor.fetchone()
        self.assertIsNone(result)
        cursor.close()
        pool.release(conn)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support proxy users yet"
    )
    def test_2407(self):
        "2407 - test heterogeneous pool with user and password specified"
        pool = test_env.get_pool(
            min=2,
            max=8,
            increment=3,
            homogeneous=False,
            getmode=oracledb.POOL_GETMODE_WAIT,
        )
        self.assertEqual(pool.homogeneous, 0)
        conn = pool.acquire()
        self.__verify_connection(pool.acquire(), test_env.get_main_user())
        conn.close()
        conn = pool.acquire(
            test_env.get_main_user(), test_env.get_main_password()
        )
        self.__verify_connection(conn, test_env.get_main_user())
        conn.close()
        conn = pool.acquire(
            test_env.get_proxy_user(), test_env.get_proxy_password()
        )
        self.__verify_connection(conn, test_env.get_proxy_user())
        conn.close()
        user_str = f"{test_env.get_main_user()}[{test_env.get_proxy_user()}]"
        conn = pool.acquire(user_str, test_env.get_main_password())
        self.assertEqual(conn.username, test_env.get_main_user())
        self.assertEqual(conn.proxy_user, test_env.get_proxy_user())
        self.__verify_connection(
            conn, test_env.get_proxy_user(), test_env.get_main_user()
        )
        conn.close()

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support proxy users yet"
    )
    def test_2408(self):
        "2408 - test heterogeneous pool without user and password specified"
        pool = test_env.get_pool(
            user="",
            password="",
            min=2,
            max=8,
            increment=3,
            getmode=oracledb.POOL_GETMODE_WAIT,
            homogeneous=False,
        )
        conn = pool.acquire(
            test_env.get_main_user(), test_env.get_main_password()
        )
        self.__verify_connection(conn, test_env.get_main_user())
        conn.close()
        conn = pool.acquire(
            test_env.get_proxy_user(), test_env.get_proxy_password()
        )
        self.__verify_connection(conn, test_env.get_proxy_user())
        conn.close()
        user_str = f"{test_env.get_main_user()}[{test_env.get_proxy_user()}]"
        conn = pool.acquire(user_str, test_env.get_main_password())
        self.__verify_connection(
            conn, test_env.get_proxy_user(), test_env.get_main_user()
        )

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support proxy users yet"
    )
    def test_2409(self):
        "2409 - test heterogeneous pool with wrong password specified"
        pool = test_env.get_pool(
            min=2,
            max=8,
            increment=3,
            getmode=oracledb.POOL_GETMODE_WAIT,
            homogeneous=False,
        )
        with self.assertRaisesFullCode("ORA-01017"):
            pool.acquire(
                test_env.get_proxy_user(), "this is the wrong password"
            )

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support tagging yet"
    )
    def test_2410(self):
        "2410 - test tagging a session"
        pool = test_env.get_pool(
            min=2, max=8, increment=3, getmode=oracledb.POOL_GETMODE_NOWAIT
        )
        tag_mst = "TIME_ZONE=MST"
        tag_utc = "TIME_ZONE=UTC"

        conn = pool.acquire()
        self.assertIsNone(conn.tag)
        pool.release(conn, tag=tag_mst)

        conn = pool.acquire()
        self.assertIsNone(conn.tag)
        conn.tag = tag_utc
        conn.close()

        conn = pool.acquire(tag=tag_mst)
        self.assertEqual(conn.tag, tag_mst)
        conn.close()

        conn = pool.acquire(tag=tag_utc)
        self.assertEqual(conn.tag, tag_utc)
        conn.close()

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support session callbacks yet",
    )
    def test_2411(self):
        "2411 - test PL/SQL session callbacks"
        if test_env.get_client_version() < (12, 2):
            self.skipTest("PL/SQL session callbacks not supported before 12.2")
        callback = "pkg_SessionCallback.TheCallback"
        pool = test_env.get_pool(
            min=2,
            max=8,
            increment=3,
            getmode=oracledb.POOL_GETMODE_NOWAIT,
            session_callback=callback,
        )
        tags = [
            "NLS_DATE_FORMAT=SIMPLE",
            "NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC",
            "NLS_DATE_FORMAT=FULL;TIME_ZONE=MST",
        ]
        actual_tags = [None, None, "NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC"]

        # truncate PL/SQL session callback log
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute("truncate table PLSQLSessionCallbacks")
        conn.close()

        # request sessions with each of the first two tags
        for tag in tags[:2]:
            conn = pool.acquire(tag=tag)
            conn.close()

        # for the last tag, use the matchanytag flag
        conn = pool.acquire(tag=tags[2], matchanytag=True)
        conn.close()

        # verify the PL/SQL session callback log is accurate
        conn = pool.acquire()
        cursor = conn.cursor()
        cursor.execute(
            """
            select RequestedTag, ActualTag
            from PLSQLSessionCallbacks
            order by FixupTimestamp
            """
        )
        results = cursor.fetchall()
        expected_results = list(zip(tags, actual_tags))
        self.assertEqual(results, expected_results)
        conn.close()

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support tagging yet"
    )
    def test_2412(self):
        "2412 - testTagging with Invalid key"
        pool = test_env.get_pool(getmode=oracledb.POOL_GETMODE_NOWAIT)
        conn = pool.acquire()
        self.assertRaises(TypeError, pool.release, conn, tag=12345)
        if test_env.get_client_version() >= (12, 2):
            with self.assertRaisesFullCode("ORA-24488"):
                pool.release(conn, tag="INVALID_TAG")

    def test_2413(self):
        "2413 - test dropping/closing a connection from the pool"
        pool = test_env.get_pool(min=1, max=5, increment=2)
        conns1 = [pool.acquire() for _ in range(2)]
        conns2 = [oracledb.connect(pool=pool) for _ in range(3)]
        self.assertEqual(pool.busy, 5)
        self.assertEqual(pool.opened, 5)

        for conn in conns1:
            pool.drop(conn)
        self.assertEqual(pool.busy, 3)
        self.assertEqual(pool.opened, 3)

        for conn in conns2:
            conn.close()
        self.assertEqual(pool.busy, 0)
        self.assertEqual(pool.opened, 3)

    def test_2414(self):
        "2414 - test to ensure pure connections are being created correctly"
        pool = test_env.get_pool(
            min=1, max=2, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        conn1 = pool.acquire()
        conn2 = pool.acquire()
        self.assertEqual(pool.opened, 2, "opened (1)")
        pool.release(conn1)
        pool.release(conn2)
        conn3 = pool.acquire(purity=oracledb.ATTR_PURITY_NEW)
        self.assertEqual(pool.opened, 2, "opened (2)")
        pool.release(conn3)

    @unittest.skipIf(
        test_env.get_is_thin(),
        "thin mode doesn't support all the pool params yet",
    )
    def test_2415(self):
        "2415 - test the reconfigure values are changed and rest unchanged"
        self.__perform_reconfigure_test("min", 5)
        self.__perform_reconfigure_test("max", 20)
        self.__perform_reconfigure_test("increment", 5)
        self.__perform_reconfigure_test("timeout", 10)
        self.__perform_reconfigure_test("stmtcachesize", 40)
        self.__perform_reconfigure_test("ping_interval", 50)
        self.__perform_reconfigure_test(
            "getmode", oracledb.POOL_GETMODE_NOWAIT
        )
        if test_env.get_client_version() >= (12, 1):
            self.__perform_reconfigure_test("max_lifetime_session", 2000)
        if test_env.get_client_version() >= (12, 2):
            self.__perform_reconfigure_test("wait_timeout", 8000)
        if test_env.get_client_version() >= (18, 3):
            self.__perform_reconfigure_test("max_sessions_per_shard", 5)
        if test_env.get_client_version() >= (19, 11):
            self.__perform_reconfigure_test("soda_metadata_cache", True)

    @unittest.skipIf(
        test_env.get_is_thin(), "thin mode doesn't support tagging yet"
    )
    def test_2417(self):
        "2417 - test that session callbacks are being called correctly"
        pool = test_env.get_pool(
            min=2,
            max=5,
            increment=1,
            session_callback=self.__callable_session_callback,
        )

        # new connection with a tag should invoke the session callback
        with pool.acquire(tag="NLS_DATE_FORMAT=SIMPLE") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(2021-05-20) from dual")
            (result,) = cursor.fetchone()
            self.assertTrue(self.session_called)

        # acquiring a connection with the same tag should not invoke the
        # session callback
        self.session_called = False
        with pool.acquire(tag="NLS_DATE_FORMAT=SIMPLE") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(2021-05-20) from dual")
            (result,) = cursor.fetchone()
            self.assertFalse(self.session_called)

        # acquiring a connection with a new tag should invoke the session
        # callback
        self.session_called = False
        with pool.acquire(tag="NLS_DATE_FORMAT=FULL;TIME_ZONE=UTC") as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            (result,) = cursor.fetchone()
            self.assertTrue(self.session_called)

        # acquiring a connection with a new tag and specifying that a
        # connection with any tag can be acquired should invoke the session
        # callback
        self.session_called = False
        with pool.acquire(
            tag="NLS_DATE_FORMAT=FULL;TIME_ZONE=MST", matchanytag=True
        ) as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            (result,) = cursor.fetchone()
            self.assertTrue(self.session_called)

        # new connection with no tag should invoke the session callback
        self.session_called = False
        with pool.acquire() as conn:
            cursor = conn.cursor()
            cursor.execute("select to_char(current_date) from dual")
            (result,) = cursor.fetchone()
            self.assertTrue(self.session_called)

    def test_2418(self):
        "2418 - test closing a pool normally with no connections checked out"
        pool = test_env.get_pool(
            min=1, max=8, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        pool.close()

    def test_2419(self):
        "2419 - test closing a pool normally with connections checked out"
        pool = test_env.get_pool(
            min=1, max=8, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        with pool.acquire():
            with self.assertRaisesFullCode("DPY-1005"):
                pool.close()

    def test_2420(self):
        "2420 - test closing a pool forcibly"
        pool = test_env.get_pool(
            min=1, max=8, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        with pool.acquire():
            pool.close(force=True)

    def test_2421(self):
        "2421 - using the pool after it is closed raises an exception"
        pool = test_env.get_pool(
            min=1, max=8, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        pool.close()
        with self.assertRaisesFullCode("DPY-1002"):
            pool.acquire()

    @unittest.skipIf(
        test_env.get_client_version() < (19, 1), "not supported on this client"
    )
    def test_2422(self):
        "2422 - using the pool beyond max limit raises an error"
        pool = test_env.get_pool(
            min=1, max=2, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        with pool.acquire(), pool.acquire():
            pool.getmode = oracledb.POOL_GETMODE_NOWAIT
            with self.assertRaisesFullCode("DPY-4005"):
                pool.acquire()

    def test_2423(self):
        "2423 - callable session callback is executed for new connections"

        class Counter:
            num_calls = 0

            @classmethod
            def session_callback(cls, conn, requested_tag):
                cls.num_calls += 1

        pool = test_env.get_pool(
            min=1,
            max=2,
            increment=1,
            session_callback=Counter.session_callback,
        )
        with pool.acquire(), pool.acquire():
            pass
        with pool.acquire(), pool.acquire():
            pass
        self.assertEqual(Counter.num_calls, 2)

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_2424(self):
        "2424 - drop the pooled connection on receiving dead connection error"
        admin_conn = test_env.get_admin_connection()
        pool = test_env.get_pool(min=2, max=2, increment=2)

        # acquire connections from the pool and kill all the sessions
        with admin_conn.cursor() as admin_cursor:
            for conn in [pool.acquire() for i in range(2)]:
                sid, serial = self.get_sid_serial(conn)
                sql = f"alter system kill session '{sid},{serial}'"
                admin_cursor.execute(sql)
                conn.close()
        self.assertEqual(pool.opened, 2)

        # when try to re-use the killed sessions error will be raised;
        # release all such connections
        for conn in [pool.acquire() for i in range(2)]:
            with conn.cursor() as cursor:
                with self.assertRaisesFullCode("DPY-4011"):
                    cursor.execute("select user from dual")
            conn.close()

        # if a free connection is available, it can be used; otherwise a new
        # connection will be created
        for conn in [pool.acquire() for i in range(2)]:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (user,) = cursor.fetchone()
                self.assertEqual(user, test_env.get_main_user().upper())
            conn.close()
        self.assertEqual(pool.opened, 2)

    def test_2425(self):
        "2425 - acquire a connection from an empty pool (min=0)"
        pool = test_env.get_pool(min=0, max=2, increment=2)
        with pool.acquire() as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (result,) = cursor.fetchone()
                self.assertEqual(result, test_env.get_main_user().upper())

    def test_2426(self):
        "2426 - get different object types from different connections"
        pool = test_env.get_pool(min=1, max=2, increment=1)
        with pool.acquire() as conn:
            typ = conn.gettype("UDT_SUBOBJECT")
            self.assertEqual(typ.name, "UDT_SUBOBJECT")
        with pool.acquire() as conn:
            typ = conn.gettype("UDT_OBJECTARRAY")
            self.assertEqual(typ.name, "UDT_OBJECTARRAY")

    def test_2427(self):
        "2427 - test creating a pool using a proxy user"
        user_str = f"{test_env.get_main_user()}[{test_env.get_proxy_user()}]"
        pool = test_env.get_pool(user=user_str)
        self.__verify_connection(
            pool.acquire(), test_env.get_proxy_user(), test_env.get_main_user()
        )

    @unittest.skipIf(test_env.get_is_drcp(), "not supported with DRCP")
    def test_2428(self):
        "2428 - test acquiring conn from pool in LIFO order"
        pool = test_env.get_pool(
            min=5, max=10, increment=1, getmode=oracledb.POOL_GETMODE_WAIT
        )
        sql = "select sys_context('userenv', 'sid') from dual"
        conns = [pool.acquire() for i in range(3)]
        sids = [conn.cursor().execute(sql).fetchone()[0] for conn in conns]

        conns[1].close()
        conns[2].close()
        conns[0].close()

        conn = pool.acquire()
        sid = conn.cursor().execute(sql).fetchone()[0]
        self.assertEqual(sid, sids[0], "not LIFO")

    def test_2429(self):
        "2429 - verify that dynamic pool cannot have an increment of zero"
        pool = test_env.get_pool(min=1, max=3, increment=0)
        self.assertEqual(pool.increment, 1)
        with pool.acquire(), pool.acquire():
            pass

    def test_2430(self):
        "2430 - verify that static pool can have an increment of zero"
        pool = test_env.get_pool(min=1, max=1, increment=0)
        self.assertEqual(pool.increment, 0)
        with pool.acquire():
            pass

    def test_2431(self):
        "2431 - verify that connection with different cclass is reused"
        cclass = "cclass2431"
        pool = test_env.get_pool(min=1, max=1)
        # ignore the first acquire which, depending on the speed with which the
        # minimum connections are created, may create a connection that is
        # discarded; instead, use the second acquire which should remain in the
        # pool
        with pool.acquire(cclass=cclass) as conn:
            pass
        with pool.acquire(cclass=cclass) as conn:
            sid_serial = self.get_sid_serial(conn)
        with pool.acquire(cclass=cclass) as conn:
            next_sid_serial = self.get_sid_serial(conn)
            self.assertEqual(next_sid_serial, sid_serial)
        self.assertEqual(pool.opened, 1)

    def test_2432(self):
        "2432 - test creating a pool invalid params"
        with self.assertRaisesFullCode("DPY-2027"):
            oracledb.create_pool(params="bad params")

    def test_2433(self):
        "2433 - test releasing and dropping an invalid connection"
        pool = test_env.get_pool()
        self.assertRaises(TypeError, pool.release, ["invalid connection"])
        self.assertRaises(TypeError, pool.drop, ["invalid connection"])

    def test_2434(self):
        "2434 - test creating a pool with invalid pool_class"
        with self.assertRaisesFullCode("DPY-2026"):
            oracledb.create_pool(pool_class=int)

    def test_2435(self):
        "2435 - test creating a pool with a subclassed connection type"

        class MyConnection(oracledb.Connection):
            pass

        pool = test_env.get_pool(connectiontype=MyConnection)
        with pool.acquire() as conn:
            self.assertIsInstance(conn, MyConnection)

    def test_2436(self):
        "2436 - test creating a pool with a subclassed pool type"

        class MyPool(oracledb.ConnectionPool):
            pass

        pool = test_env.get_pool(pool_class=MyPool)
        self.assertIsInstance(pool, MyPool)

    def test_2437(self):
        "2437 - test connectiontype with an invalid connection class"
        with self.assertRaisesFullCode("DPY-2023"):
            test_env.get_pool(connectiontype=oracledb.AsyncConnection)
        with self.assertRaisesFullCode("DPY-2023"):
            test_env.get_pool(connectiontype=int)

    @unittest.skipIf(
        test_env.get_server_version() < (12, 2), "not supported on this server"
    )
    @unittest.skipIf(
        test_env.get_client_version() < (19, 1), "not supported on this client"
    )
    def test_2438(self):
        "2438 - ensure that timed wait times out with appropriate exception"
        pool = test_env.get_pool(
            getmode=oracledb.POOL_GETMODE_TIMEDWAIT, min=0, wait_timeout=1
        )
        with self.assertRaisesFullCode("DPY-4005"):
            pool.acquire()

    @unittest.skipIf(
        test_env.get_client_version() < (18, 1), "not supported on this client"
    )
    def test_2439(self):
        "2439 - ensure call timeout is reset on connections returned by pool"
        pool = test_env.get_pool(ping_timeout=1000, ping_interval=0)
        with pool.acquire() as conn:
            self.assertEqual(conn.call_timeout, 0)
        with pool.acquire() as conn:
            self.assertEqual(conn.call_timeout, 0)

    def test_2440(self):
        "2440 - test connection with an invalid pool"
        with self.assertRaises(TypeError):
            oracledb.connect(pool="not a pool object")

    def test_2441(self):
        "2441 - test oracledb.POOL_GETMODE_FORCEGET"
        pool = test_env.get_pool(
            min=1, max=3, increment=1, getmode=oracledb.POOL_GETMODE_FORCEGET
        )
        num_conns = 10
        active_sessions = set()
        conns = [pool.acquire() for _ in range(num_conns)]
        for conn in conns:
            active_sessions.add(self.get_sid_serial(conn))
        self.assertEqual(pool.opened, num_conns)
        self.assertEqual(pool.busy, num_conns)
        self.assertEqual(len(active_sessions), num_conns)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support program yet",
    )
    def test_2442(self):
        "2442 - test passing program when creating a pool"
        sql = (
            "select program from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_create_arg("program", "newprogram", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support machine yet",
    )
    def test_2443(self):
        "2443 - test passing machine when creating a pool"
        sql = (
            "select machine from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_create_arg("machine", "newmachine", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support terminal yet",
    )
    def test_2444(self):
        "2444 - test passing terminal when creating a pool"
        sql = (
            "select terminal from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_create_arg("terminal", "newterminal", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support osuser yet",
    )
    def test_2445(self):
        "2445 - test passing osuser when creating a pool"
        sql = (
            "select osuser from v$session "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_create_arg("osuser", "newosuser", sql)

    def test_2446(self):
        "2446 - test passing driver_name when creating a pool"
        sql = (
            "select distinct client_driver from v$session_connect_info "
            "where sid = sys_context('userenv', 'sid')"
        )
        self.__verify_create_arg("driver_name", "newdriver", sql)

    @unittest.skipUnless(
        test_env.get_is_thin(),
        "thick mode doesn't support registered protocols",
    )
    def test_2447(self):
        "2447 - test register_parameter with pooled connection"
        sdu = 4096
        params = test_env.get_pool_params()
        protocol = "proto-test"
        orig_connect_string = test_env.get_connect_string()
        connect_string = f"{protocol}://{orig_connect_string}"

        def hook(passed_protocol, passed_protocol_arg, passed_params):
            self.assertEqual(passed_protocol, protocol)
            self.assertEqual(passed_protocol_arg, orig_connect_string)
            passed_params.parse_connect_string(passed_protocol_arg)
            passed_params.set(sdu=sdu)

        try:
            oracledb.register_protocol(protocol, hook)
            pool = oracledb.create_pool(dsn=connect_string, params=params)
            self.assertEqual(params.sdu, sdu)
            with pool.acquire():
                pass
            pool.close()
        finally:
            oracledb.register_protocol(protocol, None)

    def test_2448(self):
        "2448 - test create_pool() with edition"
        edition = test_env.get_edition_name()
        pool = test_env.get_pool(edition=edition)
        conn = pool.acquire()
        self.assertEqual(conn.edition, edition)

    def test_2449(self):
        "2449 - test create_pool() and get_pool() with alias"
        alias = "pool_alias_2449"
        pool = test_env.get_pool(pool_alias=alias)
        self.assertIs(pool, oracledb.get_pool(alias))
        pool.close()

    def test_2450(self):
        "2450 - test create_pool() twice with the same alias"
        alias = "pool_alias_2450"
        pool = test_env.get_pool(pool_alias=alias)
        with self.assertRaisesFullCode("DPY-2055"):
            test_env.get_pool(pool_alias=alias)
        pool.close()
        self.assertIsNone(oracledb.get_pool(alias))

    def test_2451(self):
        "2451 - test connect() with pool alias"
        alias = "pool_alias_2451"
        pool = test_env.get_pool(pool_alias=alias)
        with self.assertRaisesFullCode("DPY-2014"):
            test_env.get_connection(pool=pool, pool_alias=alias)
        with oracledb.connect(pool_alias=alias) as conn:
            with conn.cursor() as cursor:
                cursor.execute("select user from dual")
                (value,) = cursor.fetchone()
                self.assertEqual(value, test_env.get_main_user().upper())
        pool.close()
        with self.assertRaisesFullCode("DPY-2054"):
            oracledb.connect(pool_alias=alias)

    def test_2452(self):
        "2452 - test acquire() with pool alias and stmtcachesize"
        alias = "pool_2452"
        stmtcachesize = 35
        test_env.get_pool(pool_alias=alias, stmtcachesize=stmtcachesize)
        with oracledb.connect(pool_alias=alias) as conn:
            self.assertEqual(conn.stmtcachesize, stmtcachesize)
        oracledb.get_pool(alias).close()

    def test_2453(self):
        "2453 - test pool alias is case sensitive"
        alias = "pool_2458"
        test_env.get_pool(pool_alias=alias)
        self.assertIsNone(oracledb.get_pool(alias.upper()))
        with self.assertRaisesFullCode("DPY-2054"):
            test_env.get_connection(pool_alias=alias.upper())
        oracledb.get_pool(alias).close()

    def test_2454(self):
        "2454 - test pool alias with invalid types"
        aliases = [5, set(), dict(), bytearray(1)]
        for alias in aliases:
            with self.subTest(alias=alias):
                with self.assertRaises(TypeError):
                    test_env.get_pool(pool_alias=alias)

    def test_2455(self):
        "2455 - test create_pool() with parameters hook"
        pool = test_env.get_pool()
        with pool.acquire() as conn:
            orig_stmtcachesize = conn.stmtcachesize
            stmtcachesize = orig_stmtcachesize + 10
        pool.close()

        def hook(params):
            params.set(stmtcachesize=stmtcachesize)

        try:
            oracledb.register_params_hook(hook)
            pool = test_env.get_pool()
            with pool.acquire() as conn:
                self.assertEqual(conn.stmtcachesize, stmtcachesize)
            pool.close()
        finally:
            oracledb.unregister_params_hook(hook)

        pool = test_env.get_pool()
        with pool.acquire() as conn:
            self.assertEqual(conn.stmtcachesize, orig_stmtcachesize)
        pool.close()


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# pipeline.py
#
# Contains the Pipeline class used for executing multiple operations.
# -----------------------------------------------------------------------------

from typing import Any, Callable, Union

from . import __name__ as MODULE_NAME
from . import utils
from .defaults import defaults
from .fetch_info import FetchInfo
from .base_impl import PipelineImpl, PipelineOpImpl, PipelineOpResultImpl
from .enums import PipelineOpType
from .errors import _Error


class PipelineOp:
    __module__ = MODULE_NAME

    def __repr__(self):
        typ = self.__class__
        cls_name = f"{typ.__module__}.{typ.__qualname__}"
        return f"<{cls_name} of type {self.op_type.name}>"

    def _create_result(self):
        """
        Internal method used for creating a result object that is returned when
        running a pipeline.
        """
        impl = PipelineOpResultImpl(self._impl)
        result = PipelineOpResult.__new__(PipelineOpResult)
        result._operation = self
        result._impl = impl
        return result

    @property
    def arraysize(self) -> int:
        """
        Returns the array size to use when fetching all of the rows in a query.
        For all other operations the value returned is 0.
        """
        return self._impl.arraysize

    @property
    def keyword_parameters(self) -> Any:
        """
        Returns the keyword parameters to the stored procedure or function
        being called by the operation, if applicable.
        """
        return self._impl.keyword_parameters

    @property
    def name(self) -> Union[str, None]:
        """
        Returns the name of the stored procedure or function being called by
        the operation, if applicable.
        """
        return self._impl.name

    @property
    def num_rows(self) -> int:
        """
        Returns the number of rows to fetch when performing a query of a
        specific number of rows. For all operations, the value returned is 0.
        """
        return self._impl.num_rows

    @property
    def op_type(self) -> PipelineOpType:
        """
        Returns the type of operation that is taking place.
        """
        return PipelineOpType(self._impl.op_type)

    @property
    def parameters(self) -> Any:
        """
        Returns the parameters to the stored procedure or function or the
        parameters bound to the statement being executed by the operation, if
        applicable.
        """
        return self._impl.parameters

    @property
    def return_type(self) -> Any:
        """
        Returns the return type of the stored function being called by the
        operation, if applicable.
        """
        return self._impl.return_type

    @property
    def rowfactory(self) -> Union[Callable, None]:
        """
        Returns the row factory callable function to be used in a query
        executed by the operation, if applicable.
        """
        return self._impl.rowfactory

    @property
    def statement(self) -> Union[str, None]:
        """
        Returns the statement being executed by the operation, if applicable.
        """
        return self._impl.statement


class PipelineOpResult:
    __module__ = MODULE_NAME

    def __repr__(self):
        typ = self.__class__
        cls_name = f"{typ.__module__}.{typ.__qualname__}"
        return (
            f"<{cls_name} for operation of type {self.operation.op_type.name}>"
        )

    @property
    def columns(self) -> Union[list, None]:
        """
        Returns a list of FetchInfo instances containing metadata about an
        executed query, or the value None, if no fetch operation took place.
        """
        if self._impl.fetch_metadata is not None:
            return [FetchInfo._from_impl(i) for i in self._impl.fetch_metadata]

    @property
    def error(self) -> Union[_Error, None]:
        """
        Returns the error that occurred when running this operation, or the
        value None, if no error occurred.
        """
        return self._impl.error

    @property
    def operation(self) -> PipelineOp:
        """
        Returns the operation associated with the result.
        """
        return self._operation

    @property
    def return_value(self) -> Any:
        """
        Returns the return value of the called function, if a function was
        called for the operation.
        """
        return self._impl.return_value

    @property
    def rows(self) -> Union[list, None]:
        """
        Returns the rows that were fetched by the operation, if a query was
        executed.
        """
        return self._impl.rows

    @property
    def warning(self) -> Union[_Error, None]:
        """
        Returns the warning that was encountered when running this operation,
        or the value None, if no warning was encountered.
        """
        return self._impl.warning


class Pipeline:
    __module__ = MODULE_NAME

    def __repr__(self):
        typ = self.__class__
        cls_name = f"{typ.__module__}.{typ.__qualname__}"
        return f"<{cls_name} with {len(self._impl.operations)} operations>"

    def _add_op(self, op_impl):
        """
        Internal method for adding an PipelineOpImpl instance to the list of
        operations, creating an associated PipelineOp instance to correspond to
        it.
        """
        self._impl.operations.append(op_impl)
        op = PipelineOp.__new__(PipelineOp)
        op._impl = op_impl
        self._operations.append(op)
        return op

    def add_callfunc(
        self,
        name: str,
        return_type: Any,
        parameters: Union[list, tuple] = None,
        keyword_parameters: dict = None,
    ) -> PipelineOp:
        """
        Adds an operation that calls a stored function with the given
        parameters and return type. The PipelineOpResult object that is
        returned will have the "return_value" attribute populated with the
        return value of the function if the call completes successfully.
        """
        utils.verify_stored_proc_args(parameters, keyword_parameters)
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.CALL_FUNC,
            name=name,
            return_type=return_type,
            parameters=parameters,
            keyword_parameters=keyword_parameters,
        )
        return self._add_op(op_impl)

    def add_callproc(
        self,
        name: str,
        parameters: Union[list, tuple] = None,
        keyword_parameters: dict = None,
    ) -> PipelineOp:
        """
        Adds an operation that calls a stored procedure with the given
        parameters.
        """
        utils.verify_stored_proc_args(parameters, keyword_parameters)
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.CALL_PROC,
            name=name,
            parameters=parameters,
            keyword_parameters=keyword_parameters,
        )
        return self._add_op(op_impl)

    def add_commit(self) -> PipelineOp:
        """
        Adds an operation that performs a commit.
        """
        op_impl = PipelineOpImpl(op_type=PipelineOpType.COMMIT)
        return self._add_op(op_impl)

    def add_execute(
        self,
        statement: str,
        parameters: Union[list, tuple, dict] = None,
    ) -> PipelineOp:
        """
        Adds an operation that executes a statement with the given parameters.
        """
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.EXECUTE,
            statement=statement,
            parameters=parameters,
        )
        return self._add_op(op_impl)

    def add_executemany(
        self,
        statement: Union[str, None],
        parameters: Union[list, int],
    ) -> PipelineOp:
        """
        Adds an operation that executes a statement multiple times with the
        given list of parameters (or number of iterations).
        """
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.EXECUTE_MANY,
            statement=statement,
            parameters=parameters,
        )
        return self._add_op(op_impl)

    def add_fetchall(
        self,
        statement: str,
        parameters: Union[list, tuple, dict] = None,
        arraysize: int = None,
        rowfactory: Callable = None,
    ) -> PipelineOp:
        """
        Adds an operation that executes a query and returns up to the
        specified number of rows from the result set. The PipelineOpResult
        object that is returned will have the "return_value" attribute
        populated with the list of rows returned by the query.
        """
        if arraysize is None:
            arraysize = defaults.arraysize
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.FETCH_ALL,
            statement=statement,
            parameters=parameters,
            arraysize=arraysize,
            rowfactory=rowfactory,
        )
        return self._add_op(op_impl)

    def add_fetchmany(
        self,
        statement: str,
        parameters: Union[list, tuple, dict] = None,
        num_rows: int = None,
        rowfactory: Callable = None,
    ) -> PipelineOp:
        """
        Adds an operation that executes a query and returns up to the specified
        number of rows from the result set. The PipelineOpResult object that is
        returned will have the "return_value" attribute populated with the list
        of rows returned by the query.
        """
        if num_rows is None:
            num_rows = defaults.arraysize
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.FETCH_MANY,
            statement=statement,
            parameters=parameters,
            num_rows=num_rows,
            rowfactory=rowfactory,
        )
        return self._add_op(op_impl)

    def add_fetchone(
        self,
        statement: str,
        parameters: Union[list, tuple, dict] = None,
        rowfactory: Callable = None,
    ) -> PipelineOp:
        """
        Adds an operation that executes a query and returns the first row of
        the result set if one exists (or None, if no rows exist). The
        PipelineOpResult object that is returned will have the "return_value"
        attribute populated with this row if the query is performed
        successfully.
        """
        op_impl = PipelineOpImpl(
            op_type=PipelineOpType.FETCH_ONE,
            statement=statement,
            parameters=parameters,
            rowfactory=rowfactory,
        )
        return self._add_op(op_impl)

    @property
    def operations(self) -> list:
        """
        Returns the list of operations associated with the pipeline.
        """
        return self._operations


def create_pipeline() -> Pipeline:
    """
    Creates a pipeline object which can be used to process a set of operations
    against a database.
    """
    pipeline = Pipeline.__new__(Pipeline)
    pipeline._impl = PipelineImpl()
    pipeline._operations = []
    return pipeline



#------------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# decoders.pyx
#
# Cython file defining the low-level decoding routines used by the driver.
#------------------------------------------------------------------------------

cdef int decode_binary_double(const uint8_t *ptr, ssize_t num_bytes,
                              OracleDataBuffer *buffer) except -1:
    """
    Decode a binary double from raw bytes.
    """
    cdef:
        uint8_t b0, b1, b2, b3, b4, b5, b6, b7
        uint64_t high_bits, low_bits, all_bits
    b0 = ptr[0]
    b1 = ptr[1]
    b2 = ptr[2]
    b3 = ptr[3]
    b4 = ptr[4]
    b5 = ptr[5]
    b6 = ptr[6]
    b7 = ptr[7]
    if b0 & 0x80:
        b0 = b0 & 0x7f
    else:
        b0 = ~b0
        b1 = ~b1
        b2 = ~b2
        b3 = ~b3
        b4 = ~b4
        b5 = ~b5
        b6 = ~b6
        b7 = ~b7
    high_bits = b0 << 24 | b1 << 16 | b2 << 8 | b3
    low_bits = b4 << 24 | b5 << 16 | b6 << 8 | b7
    all_bits = high_bits << 32 | (low_bits & <uint64_t> 0xffffffff)
    memcpy(&buffer.as_double, &all_bits, 8)


cdef int decode_binary_float(const uint8_t *ptr, ssize_t num_bytes,
                             OracleDataBuffer *buffer) except -1:
    """
    Decode a binary float from raw bytes.
    """
    cdef:
        uint8_t b0, b1, b2, b3
        uint64_t all_bits
    b0 = ptr[0]
    b1 = ptr[1]
    b2 = ptr[2]
    b3 = ptr[3]
    if b0 & 0x80:
        b0 = b0 & 0x7f
    else:
        b0 = ~b0
        b1 = ~b1
        b2 = ~b2
        b3 = ~b3
    all_bits = (b0 << 24) | (b1 << 16) | (b2 << 8) | b3
    memcpy(&buffer.as_float, &all_bits, 4)


cdef int decode_bool(const uint8_t* ptr, ssize_t num_bytes,
                     OracleDataBuffer *buffer) except -1:
    """
    Decode a boolean value from raw bytes.
    """
    buffer.as_bool = (ptr[num_bytes - 1] == 1)


cdef int decode_date(const uint8_t* ptr, ssize_t num_bytes,
                     OracleDataBuffer *buffer) except -1:
    """
    Decode a date from the raw bytes making up the Oracle Date.
    """
    cdef OracleDate *output = &buffer.as_date
    output.year = (ptr[0] - 100) * 100 + ptr[1] - 100
    output.month = ptr[2]
    output.day = ptr[3]
    output.hour = ptr[4] - 1
    output.minute = ptr[5] - 1
    output.second = ptr[6] - 1
    if num_bytes < 11:
        output.fsecond = 0
    else:
        output.fsecond = decode_uint32be(&ptr[7]) // 1000
    if num_bytes <= 11 or ptr[11] == 0 or ptr[12] == 0:
        output.tz_hour_offset = output.tz_minute_offset = 0
    else:
        if ptr[11] & TNS_HAS_REGION_ID:
            errors._raise_err(errors.ERR_NAMED_TIMEZONE_NOT_SUPPORTED)
        output.tz_hour_offset = ptr[11] - TZ_HOUR_OFFSET
        output.tz_minute_offset = ptr[12] - TZ_MINUTE_OFFSET


cdef uint64_t decode_integer(const uint8_t* ptr, ssize_t num_bytes):
    """
    Decodes an integer from raw bytes.
    """
    if num_bytes == 1:
        return ptr[0]
    elif num_bytes == 2:
        return decode_uint16be(ptr)
    elif num_bytes == 3:
        return decode_uint24be(ptr)
    elif num_bytes == 4:
        return decode_uint32be(ptr)
    elif num_bytes == 5:
        return decode_uint40be(ptr)
    elif num_bytes == 6:
        return decode_uint48be(ptr)
    elif num_bytes == 7:
        return decode_uint56be(ptr)
    elif num_bytes == 8:
        return decode_uint64be(ptr)


cdef int decode_interval_ds(const uint8_t *ptr, ssize_t num_bytes,
                            OracleDataBuffer *buffer) except -1:
    """
    Decode an interval days to seconds from raw bytes.
    """
    cdef OracleIntervalDS *output = &buffer.as_interval_ds
    output.days = decode_uint32be(ptr) - TNS_DURATION_MID
    output.hours = ptr[4] - TNS_DURATION_OFFSET
    output.minutes = ptr[5] - TNS_DURATION_OFFSET
    output.seconds = ptr[6] - TNS_DURATION_OFFSET
    output.fseconds = decode_uint32be(&ptr[7]) - TNS_DURATION_MID


cdef int decode_interval_ym(const uint8_t *ptr, ssize_t num_bytes,
                            OracleDataBuffer *buffer) except -1:
    """
    Decode an interval years to months from raw bytes.
    """
    cdef OracleIntervalYM *output = &buffer.as_interval_ym
    output.years = decode_uint32be(ptr) - TNS_DURATION_MID
    output.months = ptr[4] - TNS_DURATION_OFFSET


cdef int decode_number(const uint8_t* ptr, ssize_t num_bytes,
                       OracleDataBuffer *buffer) except -1:
    """
    Decode a number from the raw bytes making up the Oracle number.
    """
    cdef:
        OracleNumber *output = &buffer.as_number
        uint8_t byte, digit, num_digits
        int16_t decimal_point_index
        uint8_t digits[40]
        bint is_positive
        int8_t exponent

    # the first byte is the exponent; positive numbers have the highest
    # order bit set, whereas negative numbers have the highest order bit
    # cleared and the bits inverted
    exponent = <int8_t> ptr[0]
    is_positive = (exponent & 0x80)
    if not is_positive:
        exponent = ~exponent
    exponent -= 193
    decimal_point_index = exponent * 2 + 2

    # initialize output structure
    output.is_max_negative_value = False
    output.is_integer = True
    output.num_chars = 0

    # a mantissa length of 0 implies a value of 0 (if positive) or a value
    # of -1e126 (if negative)
    if num_bytes == 1:
        if is_positive:
            output.num_chars = 1
            output.chars[0] = 48                    # zero
        else:
            output.is_max_negative_value = True
        return 0

    # check for the trailing 102 byte for negative numbers and, if present,
    # reduce the number of mantissa digits
    if not is_positive and ptr[num_bytes - 1] == 102:
        num_bytes -= 1

    # process the mantissa bytes which are the remaining bytes; each
    # mantissa byte is a base-100 digit
    num_digits = 0
    for i in range(1, num_bytes):

        # positive numbers have 1 added to them; negative numbers are
        # subtracted from the value 101
        byte = ptr[i]
        if is_positive:
            byte -= 1
        else:
            byte = 101 - byte

        # process the first digit; leading zeroes are ignored
        digit = <uint8_t> byte // 10
        if digit == 0 and num_digits == 0:
            decimal_point_index -= 1
        elif digit == 10:
            digits[num_digits] = 1
            digits[num_digits + 1] = 0
            num_digits += 2
            decimal_point_index += 1
        elif digit != 0 or i > 0:
            digits[num_digits] = digit
            num_digits += 1

        # process the second digit; trailing zeroes are ignored
        digit = byte % 10
        if digit != 0 or i < num_bytes - 1:
            digits[num_digits] = digit
            num_digits += 1

    # create string of digits for transformation to Python value
    # if negative, include the sign
    if not is_positive:
        output.chars[output.num_chars] = 45         # minus sign
        output.num_chars += 1

    # if the decimal point index is 0 or less, add the decimal point and
    # any leading zeroes that are needed
    if decimal_point_index <= 0:
        output.chars[output.num_chars] = 48         # zero
        output.chars[output.num_chars + 1] = 46     # decimal point
        output.num_chars += 2
        output.is_integer = 0
        for i in range(decimal_point_index, 0):
            output.chars[output.num_chars] = 48     # zero
            output.num_chars += 1

    # add each of the digits
    for i in range(num_digits):
        if i > 0 and i == decimal_point_index:
            output.chars[output.num_chars] = 46     # decimal point
            output.is_integer = 0
            output.num_chars += 1
        output.chars[output.num_chars] = 48 + digits[i]
        output.num_chars += 1

    # if the decimal point index exceeds the number of digits, add any
    # trailing zeroes that are needed
    if decimal_point_index > num_digits:
        for i in range(num_digits, decimal_point_index):
            output.chars[output.num_chars] = 48     # zero
            output.num_chars += 1


cdef inline uint16_t decode_uint16be(const char_type *buf):
    """
    Decodes a 16-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint16_t> buf[0] << 8) |
        (<uint16_t> buf[1])
    )


cdef inline uint16_t decode_uint16le(const char_type *buf):
    """
    Decodes a 16-bit integer in little endian order (least significant byte
    first).
    """
    return (
        (<uint16_t> buf[1] << 8) |
        (<uint16_t> buf[0])
    )


cdef inline uint32_t decode_uint24be(const char_type *buf):
    """
    Decodes a 24-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint32_t> buf[0] << 16) |
        (<uint32_t> buf[1] << 8) |
        (<uint32_t> buf[2])
    )


cdef inline uint32_t decode_uint32be(const char_type *buf):
    """
    Decodes a 32-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint32_t> buf[0] << 24) |
        (<uint32_t> buf[1] << 16) |
        (<uint32_t> buf[2] << 8) |
        (<uint32_t> buf[3])
    )


cdef inline uint64_t decode_uint40be(const char_type *buf):
    """
    Decodes a 40-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint64_t> buf[0] << 32) |
        (<uint64_t> buf[1] << 24) |
        (<uint64_t> buf[2] << 16) |
        (<uint64_t> buf[3] << 8) |
        (<uint64_t> buf[4])
    )


cdef inline uint64_t decode_uint48be(const char_type *buf):
    """
    Decodes a 48-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint64_t> buf[0] << 40) |
        (<uint64_t> buf[1] << 32) |
        (<uint64_t> buf[2] << 24) |
        (<uint64_t> buf[3] << 16) |
        (<uint64_t> buf[4] << 8) |
        (<uint64_t> buf[5])
    )


cdef inline uint64_t decode_uint56be(const char_type *buf):
    """
    Decodes a 56-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint64_t> buf[0] << 48) |
        (<uint64_t> buf[1] << 40) |
        (<uint64_t> buf[2] << 32) |
        (<uint64_t> buf[3] << 24) |
        (<uint64_t> buf[4] << 16) |
        (<uint64_t> buf[5] << 8) |
        (<uint64_t> buf[6])
    )


cdef inline uint64_t decode_uint64be(const char_type *buf):
    """
    Decodes a 64-bit integer in big endian order (most significant byte first).
    """
    return (
        (<uint64_t> buf[0] << 56) |
        (<uint64_t> buf[1] << 48) |
        (<uint64_t> buf[2] << 40) |
        (<uint64_t> buf[3] << 32) |
        (<uint64_t> buf[4] << 24) |
        (<uint64_t> buf[5] << 16) |
        (<uint64_t> buf[6] << 8) |
        (<uint64_t> buf[7])
    )



#------------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# converters.pyx
#
# Cython file defining the low-level methods for converting the intermediate
# form returned by the decoders to an appropriate Python value.
#------------------------------------------------------------------------------

cdef object convert_date_to_python(OracleDataBuffer *buffer):
    """
    Converts a DATE, TIMESTAMP, TIMESTAMP WITH LOCAL TIME ZONE or TIMESTAMP
    WITH TIMEZONE value stored in the buffer to Python datetime.datetime().
    """
    cdef:
        OracleDate *value = &buffer.as_date
        int32_t seconds
    output = cydatetime.datetime_new(value.year, value.month, value.day,
                                     value.hour, value.minute, value.second,
                                     value.fsecond, None)
    if value.tz_hour_offset != 0 or value.tz_minute_offset != 0:
        seconds = value.tz_hour_offset * 3600 + value.tz_minute_offset * 60
        output += cydatetime.timedelta_new(0, seconds, 0)
    return output


cdef object convert_interval_ds_to_python(OracleDataBuffer *buffer):
    """
    Converts an INTERVAL DAYS TO SECONDS value stored in the buffer to Python
    datetime.timedelta().
    """
    cdef:
        OracleIntervalDS *value = &buffer.as_interval_ds
        int32_t total_seconds
    total_seconds = value.hours * 60 * 60 + value.minutes * 60 + value.seconds
    return cydatetime.timedelta_new(value.days, total_seconds,
                                    value.fseconds // 1000)


cdef object convert_interval_ym_to_python(OracleDataBuffer *buffer):
    """
    Converts an INTERVAL YEARS TO MONTHS value stored in the buffer to Python
    oracledb.IntervalYM().
    """
    cdef OracleIntervalYM *value = &buffer.as_interval_ym
    return PY_TYPE_INTERVAL_YM(value.years, value.months)


cdef int convert_number_to_arrow_decimal(OracleArrowArray arrow_array,
                                         OracleDataBuffer *buffer) except -1:
    """
    Converts a NUMBER value stored in the buffer to Arrow DECIMAL128.
    """
    cdef:
        char_type c
        bint has_sign = 0
        char_type digits[39] # 38 digits + sign
        OracleNumber *value = &buffer.as_number
        uint8_t num_chars = 0, decimal_point_index = 0, allowed_max_chars = 0
        int64_t actual_scale = 0

    if value.chars[0] == 45: # minus sign
        has_sign = True

    if value.is_integer:
        if has_sign:
            allowed_max_chars = 39
        else:
            allowed_max_chars = 38
    else: # decimal point
        if has_sign:
            allowed_max_chars = 40
        else:
            allowed_max_chars = 39

    # Arrow Decimal128 can only represent values with 38 decimal digits
    if value.is_max_negative_value or value.num_chars > allowed_max_chars:
        raise ValueError("Value cannot be represented as "
                         "Arrow Decimal128")
    if value.is_integer:
        arrow_array.append_decimal(value.chars, value.num_chars)
    else:
        for i in range(value.num_chars):
            c = value.chars[i]
            # count all characters except the decimal point
            if c != 46:
                digits[num_chars] = c
                num_chars += 1
            else:
                decimal_point_index = i

        # Append any trailing zeros.
        actual_scale = num_chars - decimal_point_index
        for i in range(abs(arrow_array.scale) - actual_scale):
            digits[num_chars] = b'0'
            num_chars += 1
        arrow_array.append_decimal(digits, num_chars)



cdef int convert_number_to_arrow_double(OracleArrowArray arrow_array,
                                        OracleDataBuffer *buffer) except -1:
    """
    Converts a NUMBER value stored in the buffer to Arrow DOUBLE.
    """
    cdef OracleNumber *value = &buffer.as_number
    if value.is_max_negative_value:
        arrow_array.append_double(-1.0e126)
    else:
        arrow_array.append_double(atof(value.chars[:value.num_chars]))


cdef int convert_number_to_arrow_int64(OracleArrowArray arrow_array,
                                       OracleDataBuffer *buffer) except -1:
    """
    Converts a NUMBER value stored in the buffer to Arrow INT64.
    """
    cdef OracleNumber *value = &buffer.as_number
    arrow_array.append_int64(atoi(value.chars[:value.num_chars]))


cdef object convert_number_to_python_decimal(OracleDataBuffer *buffer):
    """
    Converts a NUMBER value stored in the buffer to Python decimal.Decimal().
    """
    cdef OracleNumber *value = &buffer.as_number
    if value.is_max_negative_value:
        return PY_TYPE_DECIMAL("-1e126")
    return PY_TYPE_DECIMAL(value.chars[:value.num_chars].decode())


cdef object convert_number_to_python_float(OracleDataBuffer *buffer):
    """
    Converts a NUMBER value stored in the buffer to Python float.
    """
    cdef OracleNumber *value = &buffer.as_number
    if value.is_max_negative_value:
        return -1.0e126
    return float(value.chars[:value.num_chars])


cdef object convert_number_to_python_int(OracleDataBuffer *buffer):
    """
    Converts a NUMBER value stored in the buffer to Python integer, if
    possible. If the value is not an integer, a float is returned instead.
    """
    cdef OracleNumber *value = &buffer.as_number
    if value.is_max_negative_value:
        return -10 ** 126
    elif value.is_integer:
        return int(value.chars[:value.num_chars])
    return float(value.chars[:value.num_chars])


cdef object convert_number_to_python_str(OracleDataBuffer *buffer):
    """
    Converts a NUMBER value stored in the buffer to Python string.
    """
    cdef OracleNumber *value = &buffer.as_number
    if value.is_max_negative_value:
        return "-1e126"
    return value.chars[:value.num_chars].decode()


cdef object convert_raw_to_python(OracleDataBuffer *buffer):
    """
    Converts a RAW or LONG RAW value stored in the buffer to Python bytes.
    """
    cdef OracleRawBytes *rb = &buffer.as_raw_bytes
    return rb.ptr[:rb.num_bytes]


cdef object convert_str_to_python(OracleDataBuffer *buffer, uint8_t csfrm,
                                  const char* encoding_errors):
    """
    Converts a CHAR, NCHAR, LONG, VARCHAR, or NVARCHAR value stored in the
    buffer to Python string.
    """
    cdef OracleRawBytes *rb = &buffer.as_raw_bytes
    if csfrm == CS_FORM_IMPLICIT:
        return rb.ptr[:rb.num_bytes].decode(ENCODING_UTF8, encoding_errors)
    return rb.ptr[:rb.num_bytes].decode(ENCODING_UTF16, encoding_errors)


cdef int convert_oracle_data_to_arrow(OracleMetadata from_metadata,
                                      OracleMetadata to_metadata,
                                      OracleData* data,
                                      OracleArrowArray arrow_array) except -1:
    """
    Converts the value stored in OracleData to Arrow format.
    """
    cdef:
        ArrowType arrow_type
        uint32_t db_type_num
        OracleRawBytes* rb
        int64_t ts

    # NULL values
    if data.is_null:
        return arrow_array.append_null()

    arrow_type = to_metadata._arrow_type
    db_type_num = from_metadata.dbtype.num
    if arrow_type == NANOARROW_TYPE_INT64:
        convert_number_to_arrow_int64(arrow_array, &data.buffer)
    elif arrow_type == NANOARROW_TYPE_DOUBLE:
        if db_type_num == DB_TYPE_NUM_NUMBER:
            convert_number_to_arrow_double(arrow_array, &data.buffer)
        else:
            arrow_array.append_double(data.buffer.as_double)
    elif arrow_type == NANOARROW_TYPE_FLOAT:
        arrow_array.append_float(data.buffer.as_float)
    elif arrow_type == NANOARROW_TYPE_BOOL:
        arrow_array.append_int64(data.buffer.as_bool)
    elif arrow_type in (
            NANOARROW_TYPE_BINARY,
            NANOARROW_TYPE_STRING,
            NANOARROW_TYPE_LARGE_BINARY,
            NANOARROW_TYPE_LARGE_STRING
    ):
        rb = &data.buffer.as_raw_bytes
        arrow_array.append_bytes(<void*> rb.ptr, rb.num_bytes)
    elif arrow_type == NANOARROW_TYPE_TIMESTAMP:
        ts = int(convert_date_to_python(&data.buffer).timestamp() *
                 arrow_array.factor)
        arrow_array.append_int64(ts)
    elif arrow_type == NANOARROW_TYPE_DECIMAL128:
        convert_number_to_arrow_decimal(arrow_array, &data.buffer)


cdef object convert_oracle_data_to_python(OracleMetadata from_metadata,
                                          OracleMetadata to_metadata,
                                          OracleData* data,
                                          const char* encoding_errors,
                                          bint from_dbobject):
    """
    Converts the value stored in OracleData to a Python object.
    """
    cdef:
        uint8_t py_type_num, ora_type_num, csfrm

    # NULL values
    if data.is_null:
        return None

    # reduce typing
    ora_type_num = from_metadata.dbtype._ora_type_num
    csfrm = from_metadata.dbtype._csfrm
    py_type_num = to_metadata._py_type_num

    # Python bytes
    if py_type_num == PY_TYPE_NUM_BYTES:

        # Oracle RAW, LONG RAW
        # Oracle CHAR, LONG and VARCHAR (bypass decode)
        if ora_type_num in (
            ORA_TYPE_NUM_CHAR,
            ORA_TYPE_NUM_LONG,
            ORA_TYPE_NUM_LONG_RAW,
            ORA_TYPE_NUM_RAW,
            ORA_TYPE_NUM_VARCHAR
        ):
            return convert_raw_to_python(&data.buffer)

    # Python string
    elif py_type_num == PY_TYPE_NUM_STR:

        # Oracle CHAR, LONG and VARCHAR
        if ora_type_num in (
            ORA_TYPE_NUM_CHAR,
            ORA_TYPE_NUM_LONG,
            ORA_TYPE_NUM_VARCHAR
        ):
            return convert_str_to_python(&data.buffer, csfrm, encoding_errors)

        # Oracle NUMBER
        elif ora_type_num == ORA_TYPE_NUM_NUMBER:
            return convert_number_to_python_str(&data.buffer)

        # Oracle BINARY_DOUBLE
        elif ora_type_num == ORA_TYPE_NUM_BINARY_DOUBLE:
            return str(data.buffer.as_double)

        # Oracle BINARY_FLOAT
        elif ora_type_num == ORA_TYPE_NUM_BINARY_FLOAT:
            return str(data.buffer.as_float)

        # Oracle DATE, TIMESTAMP (WITH (LOCAL) TIME ZONE)
        elif ora_type_num in (
            ORA_TYPE_NUM_DATE,
            ORA_TYPE_NUM_TIMESTAMP,
            ORA_TYPE_NUM_TIMESTAMP_LTZ,
            ORA_TYPE_NUM_TIMESTAMP_TZ
        ):
            return str(convert_date_to_python(&data.buffer))

        # Oracle INTERVAL DAY TO SECOND
        elif ora_type_num == ORA_TYPE_NUM_INTERVAL_DS:
            return str(convert_interval_ds_to_python(&data.buffer))

        # Oracle INTERVAL YEAR TO MONTH
        elif ora_type_num == ORA_TYPE_NUM_INTERVAL_DS:
            return str(convert_interval_ym_to_python(&data.buffer))

    # Python integer (or float if data is not an integer)
    elif py_type_num == PY_TYPE_NUM_INT:

        # Oracle BINARY_INTEGER within a DbObject
        if from_dbobject and ora_type_num == ORA_TYPE_NUM_BINARY_INTEGER:
            return data.buffer.as_integer

        # Oracle NUMBER, BINARY_INTEGER
        elif ora_type_num in (ORA_TYPE_NUM_NUMBER,
                              ORA_TYPE_NUM_BINARY_INTEGER):
            if to_metadata.dbtype._ora_type_num != ORA_TYPE_NUM_BINARY_INTEGER:
                return convert_number_to_python_int(&data.buffer)
            value = convert_number_to_python_str(&data.buffer)
            return int(PY_TYPE_DECIMAL(value))

        # Oracle CHAR, LONG, VARCHAR
        elif ora_type_num in (
            ORA_TYPE_NUM_CHAR,
            ORA_TYPE_NUM_LONG,
            ORA_TYPE_NUM_VARCHAR
        ):
            value = convert_str_to_python(&data.buffer, csfrm, encoding_errors)
            return int(PY_TYPE_DECIMAL(value))

        # Oracle BINARY_DOUBLE
        elif ora_type_num == ORA_TYPE_NUM_BINARY_DOUBLE:
            return int(PY_TYPE_DECIMAL(data.buffer.as_double))

        # Oracle BINARY_FLOAT
        elif ora_type_num == ORA_TYPE_NUM_BINARY_FLOAT:
            return int(PY_TYPE_DECIMAL(data.buffer.as_float))

    # Python decimal.Decimal
    elif py_type_num == PY_TYPE_NUM_DECIMAL:

        # Oracle NUMBER
        if ora_type_num == ORA_TYPE_NUM_NUMBER:
            return convert_number_to_python_decimal(&data.buffer)

    # Python float
    elif py_type_num == PY_TYPE_NUM_FLOAT:

        # Oracle NUMBER
        if ora_type_num == ORA_TYPE_NUM_NUMBER:
            return convert_number_to_python_float(&data.buffer)

        # Oracle BINARY_DOUBLE
        elif ora_type_num == ORA_TYPE_NUM_BINARY_DOUBLE:
            return data.buffer.as_double

        # Oracle BINARY_FLOAT
        elif ora_type_num == ORA_TYPE_NUM_BINARY_FLOAT:
            return data.buffer.as_float

        # Oracle CHAR, LONG, VARCHAR
        elif ora_type_num in (
            ORA_TYPE_NUM_CHAR,
            ORA_TYPE_NUM_LONG,
            ORA_TYPE_NUM_VARCHAR
        ):
            rb = &data.buffer.as_raw_bytes
            return float(rb.ptr[:rb.num_bytes])

    # Python datetime.datetime
    elif py_type_num == PY_TYPE_NUM_DATETIME:

        # Oracle DATE, TIMESTAMP, TIMESTAMP WITH LOCAL TIMEZONE,
        # TIMESTAMP WITH TIMEZONE
        if ora_type_num in (
            ORA_TYPE_NUM_DATE,
            ORA_TYPE_NUM_TIMESTAMP,
            ORA_TYPE_NUM_TIMESTAMP_LTZ,
            ORA_TYPE_NUM_TIMESTAMP_TZ
        ):
            return convert_date_to_python(&data.buffer)

    # Python datetime.timedelta
    elif py_type_num == PY_TYPE_NUM_TIMEDELTA:

        # Oracle INTERVAL DAY TO SECOND
        if ora_type_num == ORA_TYPE_NUM_INTERVAL_DS:
            return convert_interval_ds_to_python(&data.buffer)

    # Python oracledb.OracleIntervalYM
    elif py_type_num == PY_TYPE_NUM_ORACLE_INTERVAL_YM:

        # Oracle INTERVAL YEAR TO MONTH
        if ora_type_num == ORA_TYPE_NUM_INTERVAL_YM:
            return convert_interval_ym_to_python(&data.buffer)

    # Python boolean
    elif py_type_num == PY_TYPE_NUM_BOOL:

        # Oracle BOOLEAN
        if ora_type_num == ORA_TYPE_NUM_BOOLEAN:
            return data.buffer.as_bool

    errors._raise_err(errors.ERR_INCONSISTENT_DATATYPES,
                      input_type=from_metadata.dbtype.name,
                      output_type=to_metadata.dbtype.name)



# -----------------------------------------------------------------------------
# Copyright (c) 2020, 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

"""
3200 - Module for testing features introduced in 12.1
"""

import datetime
import unittest

import oracledb
import test_env


@unittest.skipUnless(
    test_env.get_client_version() >= (12, 1), "unsupported client"
)
class TestCase(test_env.BaseTestCase):
    def test_3200(self):
        "3200 - test executing with arraydmlrowcounts mode disabled"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [(1, "First"), (2, "Second")]
        sql = "insert into TestArrayDML (IntCol, StringCol) values (:1, :2)"
        self.cursor.executemany(sql, rows, arraydmlrowcounts=False)
        with self.assertRaisesFullCode("DPY-4006"):
            self.cursor.getarraydmlrowcounts()
        rows = [(3, "Third"), (4, "Fourth")]
        self.cursor.executemany(sql, rows)
        with self.assertRaisesFullCode("DPY-4006"):
            self.cursor.getarraydmlrowcounts()

    def test_3201(self):
        "3201 - test executing with arraydmlrowcounts mode enabled"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Fourth", 300),
            (5, "Fifth", 300),
        ]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
            arraydmlrowcounts=True,
        )
        self.conn.commit()
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 1, 1, 1])
        self.cursor.execute("select count(*) from TestArrayDML")
        (count,) = self.cursor.fetchone()
        self.assertEqual(count, len(rows))

    def test_3202(self):
        "3202 - test binding a boolean collection (in)"
        type_obj = self.conn.gettype("PKG_TESTBOOLEANS.UDT_BOOLEANLIST")
        obj = type_obj.newobject()
        obj.setelement(1, True)
        obj.extend([True, False, True, True, False, True])
        result = self.cursor.callfunc(
            "pkg_TestBooleans.TestInArrays", int, [obj]
        )
        self.assertEqual(result, 5)

    def test_3203(self):
        "3203 - test binding a boolean collection (out)"
        type_obj = self.conn.gettype("PKG_TESTBOOLEANS.UDT_BOOLEANLIST")
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestBooleans.TestOutArrays", (6, obj))
        self.assertEqual(obj.aslist(), [True, False, True, False, True, False])

    def test_3204(self):
        "3204 - test binding a PL/SQL date collection (in)"
        type_obj = self.conn.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        obj.setelement(1, datetime.datetime(2016, 2, 5))
        obj.append(datetime.datetime(2016, 2, 8, 12, 15, 30))
        obj.append(datetime.datetime(2016, 2, 12, 5, 44, 30))
        result = self.cursor.callfunc(
            "pkg_TestDateArrays.TestInArrays",
            oracledb.NUMBER,
            (2, datetime.datetime(2016, 2, 1), obj),
        )
        self.assertEqual(result, 24.75)

    def test_3205(self):
        "3205 - test binding a PL/SQL date collection (in/out)"
        type_obj = self.conn.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        obj.setelement(1, datetime.datetime(2016, 1, 1))
        obj.append(datetime.datetime(2016, 1, 7))
        obj.append(datetime.datetime(2016, 1, 13))
        obj.append(datetime.datetime(2016, 1, 19))
        self.cursor.callproc("pkg_TestDateArrays.TestInOutArrays", (4, obj))
        expected_values = [
            datetime.datetime(2016, 1, 8),
            datetime.datetime(2016, 1, 14),
            datetime.datetime(2016, 1, 20),
            datetime.datetime(2016, 1, 26),
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3206(self):
        "3206 - test binding a PL/SQL date collection (out)"
        type_obj = self.conn.gettype("PKG_TESTDATEARRAYS.UDT_DATELIST")
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestDateArrays.TestOutArrays", (3, obj))
        expected_values = [
            datetime.datetime(2002, 12, 13, 4, 48),
            datetime.datetime(2002, 12, 14, 9, 36),
            datetime.datetime(2002, 12, 15, 14, 24),
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3207(self):
        "3207 - test binding a PL/SQL number collection (in)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, 10)
        obj.extend([20, 30, 40, 50])
        result = self.cursor.callfunc(
            "pkg_TestNumberArrays.TestInArrays", int, (5, obj)
        )
        self.assertEqual(result, 155)

    def test_3208(self):
        "3208 - test binding a PL/SQL number collection (in/out)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, 5)
        obj.extend([8, 3, 2])
        self.cursor.callproc("pkg_TestNumberArrays.TestInOutArrays", (4, obj))
        self.assertEqual(obj.aslist(), [50, 80, 30, 20])

    def test_3209(self):
        "3209 - test binding a PL/SQL number collection (out)"
        type_name = "PKG_TESTNUMBERARRAYS.UDT_NUMBERLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestNumberArrays.TestOutArrays", (3, obj))
        self.assertEqual(obj.aslist(), [100, 200, 300])

    def test_3210(self):
        "3210 - test binding an array of PL/SQL records (in)"
        rec_type = self.conn.gettype("PKG_TESTRECORDS.UDT_RECORD")
        array_type = self.conn.gettype("PKG_TESTRECORDS.UDT_RECORDARRAY")
        array_obj = array_type.newobject()
        for i in range(3):
            obj = rec_type.newobject()
            obj.NUMBERVALUE = i + 1
            obj.STRINGVALUE = f"String in record #{i + 1}"
            obj.DATEVALUE = datetime.datetime(2017, i + 1, 1)
            obj.TIMESTAMPVALUE = datetime.datetime(2017, 1, i + 1)
            obj.BOOLEANVALUE = (i % 2) == 1
            obj.PLSINTEGERVALUE = i * 5
            obj.BINARYINTEGERVALUE = i * 2
            array_obj.append(obj)
        result = self.cursor.callfunc(
            "pkg_TestRecords.TestInArrays", str, [array_obj]
        )
        expected_value = (
            "udt_Record(1, 'String in record #1', "
            "to_date('2017-01-01', 'YYYY-MM-DD'), "
            "to_timestamp('2017-01-01 00:00:00', "
            "'YYYY-MM-DD HH24:MI:SS'), false, 0, 0); "
            "udt_Record(2, 'String in record #2', "
            "to_date('2017-02-01', 'YYYY-MM-DD'), "
            "to_timestamp('2017-01-02 00:00:00', "
            "'YYYY-MM-DD HH24:MI:SS'), true, 5, 2); "
            "udt_Record(3, 'String in record #3', "
            "to_date('2017-03-01', 'YYYY-MM-DD'), "
            "to_timestamp('2017-01-03 00:00:00', "
            "'YYYY-MM-DD HH24:MI:SS'), false, 10, 4)"
        )
        self.assertEqual(result, expected_value)

    def test_3211(self):
        "3211 - test binding a PL/SQL record (in)"
        type_obj = self.conn.gettype("PKG_TESTRECORDS.UDT_RECORD")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 18
        obj.STRINGVALUE = "A string in a record"
        obj.DATEVALUE = datetime.datetime(2016, 2, 15)
        obj.TIMESTAMPVALUE = datetime.datetime(2016, 2, 12, 14, 25, 36)
        obj.BOOLEANVALUE = False
        obj.PLSINTEGERVALUE = 21
        obj.BINARYINTEGERVALUE = 5
        result = self.cursor.callfunc(
            "pkg_TestRecords.GetStringRep", str, [obj]
        )
        expected_value = (
            "udt_Record(18, 'A string in a record', "
            "to_date('2016-02-15', 'YYYY-MM-DD'), "
            "to_timestamp('2016-02-12 14:25:36', "
            "'YYYY-MM-DD HH24:MI:SS'), false, 21, 5)"
        )
        self.assertEqual(result, expected_value)

    def test_3212(self):
        "3212 - test binding a PL/SQL record (out)"
        type_obj = self.conn.gettype("PKG_TESTRECORDS.UDT_RECORD")
        obj = type_obj.newobject()
        obj.NUMBERVALUE = 5
        obj.STRINGVALUE = "Test value"
        obj.DATEVALUE = datetime.datetime.today()
        obj.TIMESTAMPVALUE = datetime.datetime.today()
        obj.BOOLEANVALUE = False
        obj.PLSINTEGERVALUE = 23
        obj.BINARYINTEGERVALUE = 9
        self.cursor.callproc("pkg_TestRecords.TestOut", [obj])
        self.assertEqual(obj.NUMBERVALUE, 25)
        self.assertEqual(obj.STRINGVALUE, "String in record")
        self.assertEqual(obj.DATEVALUE, datetime.datetime(2016, 2, 16))
        self.assertEqual(
            obj.TIMESTAMPVALUE, datetime.datetime(2016, 2, 16, 18, 23, 55)
        )
        self.assertEqual(obj.BOOLEANVALUE, True)
        self.assertEqual(obj.PLSINTEGERVALUE, 45)
        self.assertEqual(obj.BINARYINTEGERVALUE, 10)

    def test_3213(self):
        "3213 - test binding a PL/SQL string collection (in)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, "First element")
        obj.setelement(2, "Second element")
        obj.setelement(3, "Third element")
        result = self.cursor.callfunc(
            "pkg_TestStringArrays.TestInArrays", int, (5, obj)
        )
        self.assertEqual(result, 45)

    def test_3214(self):
        "3214 - test binding a PL/SQL string collection (in/out)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        obj.setelement(1, "The first element")
        obj.append("The second element")
        obj.append("The third and final element")
        self.cursor.callproc("pkg_TestStringArrays.TestInOutArrays", (3, obj))
        expected_values = [
            "Converted element # 1 originally had length 17",
            "Converted element # 2 originally had length 18",
            "Converted element # 3 originally had length 27",
        ]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3215(self):
        "3215 - test binding a PL/SQL string collection (out)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestStringArrays.TestOutArrays", (4, obj))
        expected_values = [f"Test out element # {i + 1}" for i in range(4)]
        self.assertEqual(obj.aslist(), expected_values)

    def test_3216(self):
        "3216 - test binding a PL/SQL string collection (out with holes)"
        type_name = "PKG_TESTSTRINGARRAYS.UDT_STRINGLIST"
        type_obj = self.conn.gettype(type_name)
        obj = type_obj.newobject()
        self.cursor.callproc("pkg_TestStringArrays.TestIndexBy", [obj])
        self.assertEqual(obj.first(), -1048576)
        self.assertEqual(obj.last(), 8388608)
        self.assertEqual(obj.next(-576), 284)
        self.assertEqual(obj.prev(284), -576)
        self.assertEqual(obj.size(), 4)
        self.assertTrue(obj.exists(-576))
        self.assertFalse(obj.exists(-577))
        self.assertEqual(obj.getelement(284), "Third element")
        expected_list = [
            "First element",
            "Second element",
            "Third element",
            "Fourth element",
        ]
        self.assertEqual(obj.aslist(), expected_list)
        expected_dict = {
            -1048576: "First element",
            -576: "Second element",
            284: "Third element",
            8388608: "Fourth element",
        }
        self.assertEqual(obj.asdict(), expected_dict)
        obj.delete(-576)
        obj.delete(284)
        expected_list.pop(2)
        expected_list.pop(1)
        self.assertEqual(obj.aslist(), expected_list)
        expected_dict.pop(-576)
        expected_dict.pop(284)
        self.assertEqual(obj.asdict(), expected_dict)

    def test_3217(self):
        "3217 - test executing with arraydmlrowcounts with exception"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [(1, "First"), (2, "Second"), (2, "Third"), (4, "Fourth")]
        sql = "insert into TestArrayDML (IntCol,StringCol) values (:1,:2)"
        with self.assertRaisesFullCode("ORA-00001"):
            self.cursor.executemany(sql, rows, arraydmlrowcounts=True)
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1])

    def test_3218(self):
        "3218 - test executing delete statement with arraydmlrowcount mode"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Fourth", 300),
            (5, "Fifth", 300),
            (6, "Sixth", 400),
            (7, "Seventh", 400),
            (8, "Eighth", 500),
        ]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
        )
        rows = [(200,), (300,), (400,)]
        self.cursor.executemany(
            "delete from TestArrayDML where IntCol2 = :1",
            rows,
            arraydmlrowcounts=True,
        )
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 3, 2])
        self.assertEqual(self.cursor.rowcount, 6)

    def test_3219(self):
        "3219 - test executing update statement with arraydmlrowcount mode"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Fourth", 300),
            (5, "Fifth", 300),
            (6, "Sixth", 400),
            (7, "Seventh", 400),
            (8, "Eighth", 500),
        ]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
        )
        rows = [("One", 100), ("Two", 200), ("Three", 300), ("Four", 400)]
        self.cursor.executemany(
            "update TestArrayDML set StringCol = :1 where IntCol2 = :2",
            rows,
            arraydmlrowcounts=True,
        )
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 3, 2])
        self.assertEqual(self.cursor.rowcount, 7)

    def test_3220(self):
        "3220 - test getimplicitresults() returns the correct data"
        self.cursor.execute(
            """
            declare
                c1 sys_refcursor;
                c2 sys_refcursor;
            begin
                open c1 for
                    select NullableCol
                    from TestNumbers
                    where IntCol between 3 and 5;

                dbms_sql.return_result(c1);

                open c2 for
                    select NullableCol
                    from TestNumbers
                    where IntCol between 7 and 10;

                dbms_sql.return_result(c2);
            end;
            """
        )
        results = self.cursor.getimplicitresults()
        self.assertEqual(len(results), 2)
        self.assertEqual(
            [n for n, in results[0]], [2924207, None, 59797108943]
        )
        self.assertEqual(
            [n for n, in results[1]],
            [1222791080775407, None, 25004854810776297743, None],
        )

    def test_3221(self):
        "3221 - test getimplicitresults() without executing a statement"
        cursor = self.conn.cursor()
        with self.assertRaisesFullCode("DPY-1004"):
            cursor.getimplicitresults()

    def test_3222(self):
        "3222 - test executing insert with multiple distinct batch errors"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (2, "Third", 300),
            (4, "Fourth", 400),
            (5, "Fourth", 1000),
        ]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
            batcherrors=True,
            arraydmlrowcounts=True,
        )
        actual_errors = [
            (error.offset, error.full_code)
            for error in self.cursor.getbatcherrors()
        ]
        self.assertEqual(actual_errors, [(4, "ORA-01438"), (2, "ORA-00001")])
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 1, 0, 1, 0])

    def test_3223(self):
        "3223 - test batcherrors mode set to False"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [(1, "First", 100), (2, "Second", 200), (2, "Third", 300)]
        sql = """insert into TestArrayDML (IntCol, StringCol, IntCol2)
                 values (:1, :2, :3)"""
        with self.assertRaisesFullCode("ORA-00001"):
            self.cursor.executemany(sql, rows, batcherrors=False)

    def test_3224(self):
        "3224 - test executing in succession with batch error"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [
            (1, "First", 100),
            (2, "Second", 200),
            (3, "Third", 300),
            (4, "Second", 300),
            (5, "Fifth", 300),
            (6, "Sixth", 400),
            (6, "Seventh", 400),
            (8, "Eighth", 100),
        ]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
            batcherrors=True,
        )
        actual_errors = [
            (error.offset, error.full_code)
            for error in self.cursor.getbatcherrors()
        ]
        self.assertEqual(actual_errors, [(6, "ORA-00001")])
        rows = [
            (101, "First"),
            (201, "Second"),
            (3000, "Third"),
            (900, "Ninth"),
            (301, "Third"),
        ]
        self.cursor.executemany(
            "update TestArrayDML set IntCol2 = :1 where StringCol = :2",
            rows,
            arraydmlrowcounts=True,
            batcherrors=True,
        )
        actual_errors = [
            (error.offset, error.full_code)
            for error in self.cursor.getbatcherrors()
        ]
        self.assertEqual(actual_errors, [(2, "ORA-01438")])
        self.assertEqual(self.cursor.getarraydmlrowcounts(), [1, 2, 0, 0, 1])
        self.assertEqual(self.cursor.rowcount, 4)

    def test_3225(self):
        "3225 - test using implicit cursors to execute new statements"
        cursor = self.conn.cursor()
        cursor.execute(
            """
            declare
                c1 sys_refcursor;
            begin
                open c1 for
                    select NumberCol
                    from TestNumbers
                    where IntCol between 3 and 5;

                dbms_sql.return_result(c1);
            end;
            """
        )
        results = cursor.getimplicitresults()
        self.assertEqual(len(results), 1)
        self.assertEqual([n for n, in results[0]], [3.75, 5, 6.25])
        results[0].execute("select :1 from dual", [7])
        (row,) = results[0].fetchone()
        self.assertEqual(row, 7)

    def test_3226(self):
        "3226 - test batcherrors mode without any errors produced"
        self.cursor.execute("truncate table TestArrayDML")
        rows = [(1, "First", 100), (2, "Second", 200), (3, "Third", 300)]
        self.cursor.executemany(
            """
            insert into TestArrayDML (IntCol, StringCol, IntCol2)
            values (:1, :2, :3)
            """,
            rows,
            batcherrors=True,
        )
        self.assertEqual(self.cursor.getbatcherrors(), [])

    def test_3227(self):
        "3227 - test batcherrors mode with multiple executes"
        self.cursor.execute("truncate table TestArrayDML")
        rows_1 = [
            (1, "Value 1", 100),
            (2, "Value 2", 200),
            (2, "Value 2", 200),
        ]
        rows_2 = [
            (3, "Value 3", 300),
            (3, "Value 3", 300),
            (4, "Value 4", 400),
        ]
        sql = """
                insert into TestArrayDML (IntCol, StringCol, IntCol2)
                values (:1, :2, :3)"""
        self.cursor.executemany(sql, rows_1, batcherrors=True)
        actual_errors = [
            (error.offset, error.full_code)
            for error in self.cursor.getbatcherrors()
        ]
        self.assertEqual(actual_errors, [(2, "ORA-00001")])
        self.cursor.executemany(sql, rows_2, batcherrors=True)
        actual_errors = [
            (error.offset, error.full_code)
            for error in self.cursor.getbatcherrors()
        ]
        self.assertEqual(actual_errors, [(1, "ORA-00001")])

    def test_3228(self):
        "3228 - test %ROWTYPE record type"
        type_obj = self.conn.gettype("TESTTEMPTABLE%ROWTYPE")
        self.assertEqual(type_obj.attributes[3].name, "NUMBERCOL")

    def test_3229(self):
        "3229 - test collection of %ROWTYPE record type"
        type_name = "PKG_TESTBINDOBJECT.UDT_COLLECTIONROWTYPE"
        type_obj = self.conn.gettype(type_name)
        self.assertEqual(type_obj.element_type.attributes[3].name, "NUMBERCOL")

    def test_3230(self):
        "3230 - enabling batcherrors parameter with PL/SQL"
        with self.assertRaisesFullCode("DPY-2040"):
            self.cursor.executemany("begin null; end;", 30, batcherrors=True)

    def test_3231(self):
        "3231 - enabling arraydmlrowcountsbatcherrors parameter with PL/SQL"
        with self.assertRaisesFullCode("DPY-2040"):
            self.cursor.executemany(
                "begin null; end;", 31, arraydmlrowcounts=True
            )

    def test_3232(self):
        "3232 - fetch implicit cursors after closing connection"
        conn = test_env.get_connection()
        cursor = conn.cursor()
        cursor.execute(
            """
            declare
                c1 sys_refcursor;
                c2 sys_refcursor;
            begin
                open c1 for
                    select NullableCol
                    from TestNumbers;

                dbms_sql.return_result(c1);

                open c2 for
                    select NullableCol
                    from TestNumbers;

                dbms_sql.return_result(c2);
            end;
            """
        )
        cursor1, cursor2 = cursor.getimplicitresults()
        conn.close()
        with self.assertRaisesFullCode("DPY-1001"):
            cursor1.fetchall()
        with self.assertRaisesFullCode("DPY-1001"):
            cursor2.fetchall()

    def test_3233(self):
        "3233 - fetch implicit cursors after closing parent cursor"
        cursor = self.conn.cursor()
        cursor.execute(
            """
            declare
                c1 sys_refcursor;
                c2 sys_refcursor;
            begin
                open c1 for
                    select NullableCol
                    from TestNumbers
                    where IntCol between 3 and 5;

                dbms_sql.return_result(c1);

                open c2 for
                    select NullableCol
                    from TestNumbers
                    where IntCol between 7 and 10;

                dbms_sql.return_result(c2);
            end;
            """
        )
        cursor1, cursor2 = cursor.getimplicitresults()
        cursor.close()
        if self.conn.thin:
            self.assertEqual(
                [n for n, in cursor1], [2924207, None, 59797108943]
            )
            self.assertEqual(
                [n for n, in cursor2],
                [1222791080775407, None, 25004854810776297743, None],
            )
        else:
            with self.assertRaisesFullCode("DPI-1039"):
                cursor1.fetchall()
            with self.assertRaisesFullCode("DPI-1039"):
                cursor1.fetchall()

    def test_3234(self):
        "3234 - test PL/SQL record metadata"
        rec_type = self.conn.gettype("PKG_TESTRECORDS.UDT_RECORD")
        expected_metadata = [
            ("NUMBERVALUE", oracledb.DB_TYPE_NUMBER, 0, -127, None),
            ("STRINGVALUE", oracledb.DB_TYPE_VARCHAR, None, None, 30),
            ("DATEVALUE", oracledb.DB_TYPE_DATE, None, None, None),
            ("TIMESTAMPVALUE", oracledb.DB_TYPE_TIMESTAMP, None, None, None),
            ("BOOLEANVALUE", oracledb.DB_TYPE_BOOLEAN, None, None, None),
            (
                "PLSINTEGERVALUE",
                oracledb.DB_TYPE_BINARY_INTEGER,
                None,
                None,
                None,
            ),
            (
                "BINARYINTEGERVALUE",
                oracledb.DB_TYPE_BINARY_INTEGER,
                None,
                None,
                None,
            ),
        ]
        actual_metadata = [
            (attr.name, attr.type, attr.precision, attr.scale, attr.max_size)
            for attr in rec_type.attributes
        ]
        self.assertEqual(actual_metadata, expected_metadata)


if __name__ == "__main__":
    test_env.run_test_cases()



# -----------------------------------------------------------------------------
# Copyright (c) 2024, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# sparse_vector.py
#
# Contains the SparseVector class which stores information about a sparse
# vector. Sparse vectors are available in Oracle Database 23.6 and higher.
# -----------------------------------------------------------------------------

import array
from typing import Union

from .base_impl import get_array_type_code_uint32, SparseVectorImpl
from . import __name__ as MODULE_NAME

ARRAY_TYPE_CODE_UINT32 = get_array_type_code_uint32()


class SparseVector:
    """
    Provides information about sparse vectors.
    """

    __module__ = MODULE_NAME

    def __init__(
        self,
        num_dimensions: int,
        indices: Union[list, array.array],
        values: Union[list, array.array],
    ):
        if (
            not isinstance(indices, array.array)
            or indices.typecode != ARRAY_TYPE_CODE_UINT32
        ):
            indices = array.array(ARRAY_TYPE_CODE_UINT32, indices)
        if not isinstance(values, array.array):
            values = array.array("d", values)
        if len(indices) != len(values):
            raise TypeError("indices and values must be of the same length!")
        self._impl = SparseVectorImpl.from_values(
            num_dimensions, indices, values
        )

    def __repr__(self):
        return (
            f"{MODULE_NAME}.{self.__class__.__name__}({self.num_dimensions}, "
            f"{self.indices}, {self.values})"
        )

    def __str__(self):
        return (
            f"[{self.num_dimensions}, {list(self.indices)}, "
            f"{list(self.values)}]"
        )

    @classmethod
    def _from_impl(cls, impl):
        vector = cls.__new__(cls)
        vector._impl = impl
        return vector

    @property
    def indices(self) -> array.array:
        """
        Returns the indices (zero-based) of non-zero values in the vector.
        """
        return self._impl.indices

    @property
    def num_dimensions(self) -> int:
        """
        Returns the number of dimensions contained in the vector.
        """
        return self._impl.num_dimensions

    @property
    def values(self) -> array.array:
        """
        Returns the non-zero values stored in the vector.
        """
        return self._impl.values



# -----------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# buffer.py
#
# Implements the Buffer class as documented in DataFrame API
# -----------------------------------------------------------------------------

from typing import Tuple

from .protocol import (
    Buffer,
    DlpackDeviceType,
)


class OracleColumnBuffer(Buffer):
    """
    OracleColumnBuffer represents a contiguous memory buffer in the DataFrame
    Interchange Protocol. It provides access to raw binary data that backs
    various components of the data frame such as column values, validity masks
    and offsets for variable-length data types.
    """

    def __init__(self, buffer_type, size_in_bytes, address) -> None:
        self.buffer_type = buffer_type
        self.size_in_bytes = size_in_bytes
        self.address = address

    def __dlpack__(self):
        """
        Represent this structure as a DLPack interface.
        """
        raise NotImplementedError("__dlpack__")

    def __dlpack_device__(self) -> Tuple[DlpackDeviceType, None]:
        """
        Device type and device ID for where the data
        in the buffer resides
        """
        return (DlpackDeviceType.CPU, None)

    def __repr__(self) -> str:
        device = self.__dlpack_device__()[0].name
        return (
            f"OracleColumnBuffer(bufsize={self.bufsize}, "
            f"ptr={self.ptr}, type={self.buffer_type}, device={device!r})"
        )

    @property
    def bufsize(self) -> int:
        """
        Returns the total size of buffer in bytes.
        """
        return self.size_in_bytes

    @property
    def ptr(self) -> int:
        """
        Returns the memory address of the buffer.
        """
        return self.address



#------------------------------------------------------------------------------
# Copyright (c) 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# aq_enq.pyx
#
# Cython file defining the messages that are sent to the database and the
# responses that are received by the client for enqueuing an AQ message
# (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

@cython.final
cdef class AqEnqMessage(AqBaseMessage):
    cdef:
        ThinMsgPropsImpl props_impl

    cdef int _initialize_hook(self) except -1:
        """
        perform initialization
        """
        self.function_code = TNS_FUNC_AQ_ENQ

    cdef int _process_return_parameters(self, ReadBuffer buf) except -1:
        """
        Process the return parameters for the AQ enqueue request.
        """
        cdef const char_type *ptr = buf._get_raw(TNS_AQ_MESSAGE_ID_LENGTH)
        self.props_impl.msgid = ptr[:TNS_AQ_MESSAGE_ID_LENGTH]
        buf.skip_ub2()                          # extensions length

    cdef int _write_message(self, WriteBuffer buf) except -1:
        """
        Write message to the network buffers.
        """
        cdef:
            bytes queue_name_bytes
            bytes correlation_bytes
            bytes exceptionq_bytes
            int enq_flags

        self._write_function_code(buf)
        queue_name_bytes = self.queue_impl.name.encode()
        buf.write_uint8(1)                      # queue name (pointer)
        buf.write_ub4(len(queue_name_bytes))    # queue name length
        self._write_msg_props(buf, self.props_impl)
        buf.write_uint8(0)                      # recipients (pointer)
        buf.write_ub4(0)                        # number of key/value pairs
        buf.write_ub4(self.enq_options_impl.visibility)
        buf.write_uint8(0)                      # relative message id
        buf.write_ub4(0)                        # relative message length
        buf.write_ub4(0)                        # sequence deviation
        buf.write_uint8(1)                      # TOID of payload (pointer)
        buf.write_ub4(16)                       # TOID of payload length
        buf.write_ub2(TNS_AQ_MESSAGE_VERSION)
        if self.queue_impl.is_json:
            buf.write_uint8(0)                  # payload (pointer)
            buf.write_uint8(0)                  # RAW payload (pointer)
            buf.write_ub4(0)                    # RAW payload length
        elif self.queue_impl.payload_type is not None:
            buf.write_uint8(1)                  # payload (pointer)
            buf.write_uint8(0)                  # RAW payload (pointer)
            buf.write_ub4(0)                    # RAW payload (length)
        else:
            buf.write_uint8(0)                  # payload (pointer)
            buf.write_uint8(1)                  # RAW payload (pointer)
            buf.write_ub4(len(self.props_impl.payload_obj))
        buf.write_uint8(1)                      # return message id (pointer)
        buf.write_ub4(TNS_AQ_MESSAGE_ID_LENGTH) # return message id length
        enq_flags = 0
        if self.enq_options_impl.delivery_mode == TNS_AQ_MSG_BUFFERED:
            enq_flags |= TNS_KPD_AQ_BUFMSG
        buf.write_ub4(enq_flags)                # enqueue flags
        buf.write_uint8(0)                      # extensions 1 (pointer)
        buf.write_ub4(0)                        # number of extensions 1
        buf.write_uint8(0)                      # extensions 2 (pointer)
        buf.write_ub4(0)                        # number of extensions 2
        buf.write_uint8(0)                      # source sequence number
        buf.write_ub4(0)                        # source sequence length
        buf.write_uint8(0)                      # max sequence number
        buf.write_ub4(0)                        # max sequence length
        buf.write_uint8(0)                      # output ack length
        buf.write_uint8(0)                      # correlation (pointer)
        buf.write_ub4(0)                        # correlation length
        buf.write_uint8(0)                      # sender name (pointer)
        buf.write_ub4(0)                        # sender name length
        buf.write_uint8(0)                      # sender address (pointer)
        buf.write_ub4(0)                        # sender address length
        buf.write_uint8(0)                      # sender charset id (pointer)
        buf.write_uint8(0)                      # sender ncharset id (pointer)
        if buf._caps.ttc_field_version >= TNS_CCAP_FIELD_VERSION_20_1:
            if self.queue_impl.is_json:
                buf.write_uint8(1)              # JSON payload (pointer)
            else:
                buf.write_uint8(0)              # JSON payload (pointer)

        buf.write_bytes_with_length(queue_name_bytes)
        buf.write_bytes(self.queue_impl.payload_toid)
        self._write_payload(buf, self.props_impl)



#------------------------------------------------------------------------------
# Copyright (c) 2024, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# metadata.pyx
#
# Cython file defining the OracleMetadata class (embedded in base_impl.pyx).
#------------------------------------------------------------------------------

@cython.freelist(30)
cdef class OracleMetadata:

    cdef int _finalize_init(self) except -1:
        """
        Internal method that finalizes the initialization of metadata by
        setting the buffer size, max size and default Python type (if they have
        not already been set).
        """
        if self.dbtype.default_size == 0:
            self.max_size = 0
            self.buffer_size = self.dbtype._buffer_size_factor
        else:
            if self.max_size == 0:
                self.max_size = self.dbtype.default_size
            self.buffer_size = self.max_size * self.dbtype._buffer_size_factor
        if self._py_type_num == 0:
            if self.dbtype._ora_type_num != ORA_TYPE_NUM_NUMBER:
                self._py_type_num = self.dbtype._default_py_type_num
            else:
                if self.scale == 0 or \
                        (self.scale == -127 and self.precision == 0):
                    self._py_type_num = PY_TYPE_NUM_INT
                else:
                    self._py_type_num = PY_TYPE_NUM_FLOAT

    cdef int _set_arrow_type(self) except -1:
        """
        Determine the arrow type to use for the data.
        """
        cdef:
            uint8_t py_type_num = self._py_type_num
            uint32_t db_type_num = self.dbtype.num
        if db_type_num == DB_TYPE_NUM_NUMBER:
            if py_type_num == PY_TYPE_NUM_DECIMAL and self.precision > 0:
                self._arrow_type = NANOARROW_TYPE_DECIMAL128
            elif py_type_num == PY_TYPE_NUM_STR:
                self._arrow_type = NANOARROW_TYPE_STRING
            elif py_type_num == PY_TYPE_NUM_INT and self.scale == 0 \
                    and self.precision <= 18:
                self._arrow_type = NANOARROW_TYPE_INT64
            else:
                self._arrow_type = NANOARROW_TYPE_DOUBLE
        elif db_type_num in (DB_TYPE_NUM_CHAR, DB_TYPE_NUM_VARCHAR):
            self._arrow_type = NANOARROW_TYPE_STRING
        elif db_type_num == DB_TYPE_NUM_BINARY_FLOAT:
            self._arrow_type = NANOARROW_TYPE_FLOAT
        elif db_type_num == DB_TYPE_NUM_BINARY_DOUBLE:
            self._arrow_type = NANOARROW_TYPE_DOUBLE
        elif db_type_num == DB_TYPE_NUM_BOOLEAN:
            self._arrow_type = NANOARROW_TYPE_BOOL
        elif db_type_num in (DB_TYPE_NUM_DATE,
                             DB_TYPE_NUM_TIMESTAMP,
                             DB_TYPE_NUM_TIMESTAMP_LTZ,
                             DB_TYPE_NUM_TIMESTAMP_TZ):
            self._arrow_type = NANOARROW_TYPE_TIMESTAMP
        elif db_type_num == DB_TYPE_NUM_LONG_RAW:
            self._arrow_type = NANOARROW_TYPE_LARGE_BINARY
        elif db_type_num == DB_TYPE_NUM_LONG_VARCHAR:
            self._arrow_type = NANOARROW_TYPE_LARGE_STRING
        elif db_type_num == DB_TYPE_NUM_RAW:
            self._arrow_type = NANOARROW_TYPE_BINARY
        else:
            errors._raise_err(errors.ERR_ARROW_UNSUPPORTED_DATA_TYPE,
                              db_type_name=self.dbtype.name)

    cdef OracleMetadata copy(self):
        """
        Create a copy of the metadata and return it.
        """
        cdef OracleMetadata metadata = OracleMetadata.__new__(OracleMetadata)
        metadata.name = self.name
        metadata.dbtype = self.dbtype
        metadata.objtype = self.objtype
        metadata.precision = self.precision
        metadata.scale = self.scale
        metadata.max_size = self.max_size
        metadata.nulls_allowed = self.nulls_allowed
        metadata.is_json = self.is_json
        metadata.is_oson = self.is_oson
        metadata.domain_schema = self.domain_schema
        metadata.domain_name = self.domain_name
        metadata.annotations = self.annotations
        metadata.vector_dimensions = self.vector_dimensions
        metadata.vector_format = self.vector_format
        metadata.vector_flags = self.vector_flags
        return metadata

    @staticmethod
    cdef OracleMetadata from_type(object typ):
        """
        Returns a new OracleMetadata instance with attributes set according to
        the Python type.
        """
        cdef:
            OracleMetadata metadata = OracleMetadata.__new__(OracleMetadata)
            ApiType apitype
        if isinstance(typ, DbType):
            metadata.dbtype = typ
        elif isinstance(typ, ApiType):
            apitype = typ
            metadata.dbtype = apitype.dbtypes[0]
        elif isinstance(typ, PY_TYPE_DB_OBJECT_TYPE):
            metadata.dbtype = DB_TYPE_OBJECT
            metadata.objtype = typ._impl
        elif not isinstance(typ, type):
            errors._raise_err(errors.ERR_EXPECTING_TYPE)
        elif typ is int:
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_INT
        elif typ is float:
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_FLOAT
        elif typ is str:
            metadata.dbtype = DB_TYPE_VARCHAR
        elif typ is bytes:
            metadata.dbtype = DB_TYPE_RAW
        elif typ is PY_TYPE_DECIMAL:
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_DECIMAL
        elif typ is PY_TYPE_BOOL:
            metadata.dbtype = DB_TYPE_BOOLEAN
        elif typ is PY_TYPE_DATE:
            metadata.dbtype = DB_TYPE_DATE
        elif typ is PY_TYPE_DATETIME:
            metadata.dbtype = DB_TYPE_TIMESTAMP
        elif typ is PY_TYPE_TIMEDELTA:
            metadata.dbtype = DB_TYPE_INTERVAL_DS
        else:
            errors._raise_err(errors.ERR_PYTHON_TYPE_NOT_SUPPORTED, typ=typ)
        return metadata

    @staticmethod
    cdef OracleMetadata from_value(object value):
        """
        Returns a new OracleMetadata instance with attributes set according to
        the Python type.
        """
        cdef OracleMetadata metadata = OracleMetadata.__new__(OracleMetadata)
        if value is None:
            metadata.dbtype = DB_TYPE_VARCHAR
            metadata.max_size = 1
        elif isinstance(value, PY_TYPE_BOOL):
            metadata.dbtype = DB_TYPE_BOOLEAN
        elif isinstance(value, str):
            metadata.dbtype = DB_TYPE_VARCHAR
            metadata.max_size = <uint32_t> len((<str> value).encode())
        elif isinstance(value, bytes):
            metadata.max_size = <uint32_t> len(value)
            metadata.dbtype = DB_TYPE_RAW
        elif isinstance(value, int):
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_INT
        elif isinstance(value, float):
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_FLOAT
        elif isinstance(value, PY_TYPE_DECIMAL):
            metadata.dbtype = DB_TYPE_NUMBER
            metadata._py_type_num = PY_TYPE_NUM_DECIMAL
        elif isinstance(value, (PY_TYPE_DATE, PY_TYPE_DATETIME)):
            metadata.dbtype = DB_TYPE_DATE
        elif isinstance(value, PY_TYPE_TIMEDELTA):
            metadata.dbtype = DB_TYPE_INTERVAL_DS
        elif isinstance(value, PY_TYPE_DB_OBJECT):
            metadata.dbtype = DB_TYPE_OBJECT
            metadata.objtype = value.type._impl
        elif isinstance(value, (PY_TYPE_LOB, PY_TYPE_ASYNC_LOB)):
            metadata.dbtype = value.type
        elif isinstance(value, (PY_TYPE_CURSOR, PY_TYPE_ASYNC_CURSOR)):
            metadata.dbtype = DB_TYPE_CURSOR
        elif isinstance(value, (array.array, PY_TYPE_SPARSE_VECTOR)):
            metadata.dbtype = DB_TYPE_VECTOR
        elif isinstance(value, PY_TYPE_INTERVAL_YM):
            metadata.dbtype = DB_TYPE_INTERVAL_YM
        else:
            errors._raise_err(errors.ERR_PYTHON_VALUE_NOT_SUPPORTED,
                              type_name=type(value).__name__)
        return metadata



#------------------------------------------------------------------------------
# Copyright (c) 2020, 2025, Oracle and/or its affiliates.
#
# This software is dual-licensed to you under the Universal Permissive License
# (UPL) 1.0 as shown at https://oss.oracle.com/licenses/upl and Apache License
# 2.0 as shown at http://www.apache.org/licenses/LICENSE-2.0. You may choose
# either license.
#
# If you elect to accept the software under the Apache License, Version 2.0,
# the following applies:
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# utils.pyx
#
# Cython file defining utility classes and methods (embedded in thin_impl.pyx).
#------------------------------------------------------------------------------

class OutOfPackets(Exception):
    pass

class MarkerDetected(Exception):
    pass

class ConnectConstants:

    def __init__(self):
        self.pid = str(os.getpid())
        pattern = r"(?P<major_num>\d+)\.(?P<minor_num>\d+)\.(?P<patch_num>\d+)"
        match_dict = re.match(pattern, DRIVER_VERSION)
        major_num = int(match_dict["major_num"])
        minor_num = int(match_dict["minor_num"])
        patch_num = int(match_dict["patch_num"])
        self.full_version_num = \
                major_num << 24 | minor_num << 20 | patch_num << 12


cdef int _convert_base64(char_type *buf, long value, int size, int offset):
    """
    Converts every 6 bits into a character, from left to right. This is
    similar to ordinary base64 encoding with a few differences and written
    here for performance.
    """
    cdef int i
    for i in range(size):
        buf[offset + size - i - 1] = TNS_BASE64_ALPHABET_ARRAY[value & 0x3f]
        value = value >> 6
    return offset + size


cdef object _encode_rowid(Rowid *rowid):
    """
    Converts the rowid structure into an encoded string, if the rowid structure
    contains valid data; otherwise, it returns None.
    """
    cdef:
        char_type buf[TNS_MAX_ROWID_LENGTH]
        int offset = 0
    if rowid.rba != 0 or rowid.partition_id != 0 or rowid.block_num != 0 \
            or rowid.slot_num != 0:
        offset = _convert_base64(buf, rowid.rba, 6, offset)
        offset = _convert_base64(buf, rowid.partition_id, 3, offset)
        offset = _convert_base64(buf, rowid.block_num, 6, offset)
        offset = _convert_base64(buf, rowid.slot_num, 3, offset)
        return buf[:TNS_MAX_ROWID_LENGTH].decode()


cdef str _get_connect_data(Description description, str connection_id, ConnectParamsImpl params):
    """
    Return the connect data required by the listener in order to connect.
    """
    cid = f"(PROGRAM={params.program})" + \
          f"(HOST={params.machine})" + \
          f"(USER={params.osuser})"
    if description.connection_id_prefix:
        description.connection_id = description.connection_id_prefix + \
                connection_id
    else:
        description.connection_id = connection_id
    return description.build_connect_string(cid)


cdef int _check_cryptography() except -1:
    """
    Checks to see that the cryptography package was imported successfully.
    """
    if CRYPTOGRAPHY_IMPORT_ERROR is not None:
        errors._raise_err(errors.ERR_NO_CRYPTOGRAPHY_PACKAGE,
                          str(CRYPTOGRAPHY_IMPORT_ERROR))


def init_thin_impl(package):
    """
    Initializes globals after the package has been completely initialized. This
    is to avoid circular imports and eliminate the need for global lookups.
    """
    global _connect_constants, errors, exceptions
    _connect_constants = ConnectConstants()
    errors = package.errors
    exceptions = package.exceptions
